﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CashCalc.Base
{
    /// <summary>
    /// Base abstract class for all (serializable) accounts.
    /// </summary>
    [Serializable]
    public abstract class Account : IComparable<Account>
    {
        /// <summary>
        /// Account title.
        /// </summary>
        public virtual string Title { get; set; }

        /// <summary>
        /// Computes the expected account balance and its standard deviation at the specified date.
        /// </summary>
        /// <param name="date">Date of the forecast.</param>
        /// <returns>
        /// Returns the AmountWithDeviation structure representing the expected account balance 
        /// and its standard deviation at the specified date.
        /// </returns>
        public abstract AmountWithDeviation ExpectedBalance(DateTime date);

        /// <summary>
        /// Returns the explanation of the total change of the account balance at the specified date.
        /// </summary>
        /// <param name="date">Date of the account balance change.</param>
        /// <returns>
        /// Returns the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </returns>
        public abstract AmountExplanation Explain(DateTime date);

        /// <summary>
        /// Returns the zeroized day forecast, mapping every day in the given range to the AmountWithDeviation.Zero 
        /// instance representing the zero expected account balance and zero standard deviation.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast.</param>
        /// <param name="lastDay">Last day of the day forecast.</param>
        /// <returns>Zeroized day forecast.</returns>
        public static SortedList<DateTime, AmountWithDeviation> GetZeroDayForecast(DateTime firstDay, DateTime lastDay)
        {
            firstDay = firstDay.Date;
            lastDay = lastDay.Date;
            if (firstDay >= lastDay)
            {
                throw new ArgumentException("First day must be strictly before the last day");
            }
            SortedList<DateTime, AmountWithDeviation> dayForecast =
                new SortedList<DateTime, AmountWithDeviation>((int)(lastDay - firstDay).TotalDays + 1);
            DateTime date = firstDay;
            while (date <= lastDay)
            {
                dayForecast[date] = AmountWithDeviation.Zero;
                date = date.AddTicks(TimeSpan.TicksPerDay);
            }
            return dayForecast;
        }

        /// <summary>
        /// Returns the zeroized month forecast, mapping every month in the given range to the AmountWithDeviation.Zero 
        /// instance representing the zero expected account balance and zero standard deviation.
        /// </summary>
        /// <param name="firstMonth">First month of the day forecast.</param>
        /// <param name="lastMonth">Last month of the day forecast.</param>
        /// <returns>Zeroized day forecast.</returns>
        public static SortedList<DateTime, AmountWithDeviation> GetZeroMonthForecast(DateTime firstMonth, DateTime lastMonth)
        {
            firstMonth = new DateTime(firstMonth.Year, firstMonth.Month, 1);
            lastMonth = new DateTime(lastMonth.Year, lastMonth.Month, 1);
            if (firstMonth >= lastMonth)
            {
                throw new ArgumentException("First month must be strictly before the last month");
            }
            SortedList<DateTime, AmountWithDeviation> monthForecast =
                new SortedList<DateTime, AmountWithDeviation>
                    (12 * (lastMonth.Year - firstMonth.Year) + lastMonth.Month - firstMonth.Month + 1);
            DateTime month = firstMonth;
            while (month <= lastMonth)
            {
                monthForecast[month] = AmountWithDeviation.Zero;
                month = month.AddMonths(1);
            }
            return monthForecast;
        }

        /// <summary>
        /// Computes the zeroized explanation of the day forecast, mapping every day in the given range 
        /// to the AmountExplanation.Zero instance representing the explanation of the total zero change 
        /// of the account balance at the specified date.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast explanation.</param>
        /// <param name="lastDay">Last day of the day forecast explanation.</param>
        /// <returns>Zeroized day forecast explanation.</returns>
        public static SortedList<DateTime, AmountExplanation> GetZeroDayExplanation(DateTime firstDay, DateTime lastDay)
        {
            firstDay = firstDay.Date;
            lastDay = lastDay.Date;
            if (firstDay >= lastDay)
            {
                throw new ArgumentException("First day must be strictly before the last day");
            }
            SortedList<DateTime, AmountExplanation> dayForecastExplanation =
                new SortedList<DateTime, AmountExplanation>((int)(lastDay - firstDay).TotalDays + 1);
            DateTime date = firstDay;
            while (date <= lastDay)
            {
                dayForecastExplanation[date] = AmountExplanation.Zero;
                date = date.AddTicks(TimeSpan.TicksPerDay);
            }
            return dayForecastExplanation;
        }

        /// <summary>
        /// Returns the zeroized explanation of the month forecast, mapping every month in the given range 
        /// to the AmountExplanation.Zero instance representing the explanation of the total zero change 
        /// of the account balance at the specified month.
        /// </summary>
        /// <param name="firstMonth">First month of the month forecast explanation.</param>
        /// <param name="lastMonth">Last month of the month forecast explanation.</param>
        /// <returns>Zeroized month forecast explanation.</returns>
        public static SortedList<DateTime, AmountExplanation> GetZeroMonthExplanation(DateTime firstMonth, DateTime lastMonth)
        {
            firstMonth = new DateTime(firstMonth.Year, firstMonth.Month, 1);
            lastMonth = new DateTime(lastMonth.Year, lastMonth.Month, 1);
            if (firstMonth >= lastMonth)
            {
                throw new ArgumentException("First month must be strictly before the last month");
            }
            SortedList<DateTime, AmountExplanation> monthForecastExplanation =
                new SortedList<DateTime, AmountExplanation>
                    (12 * (lastMonth.Year - firstMonth.Year) + lastMonth.Month - firstMonth.Month + 1);
            DateTime month = firstMonth;
            while (month <= lastMonth)
            {
                monthForecastExplanation[month] = AmountExplanation.Zero;
                month = month.AddMonths(1);
            }
            return monthForecastExplanation;
        }

        /// <summary>
        /// Computes the day forecast, mapping every day in the given range to the AmountWithDeviation 
        /// structure representing the expected account balance and its standard deviation.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast.</param>
        /// <param name="lastDay">Last day of the day forecast.</param>
        /// <param name="dayForecast">
        /// Day forecast to be computed, mapping every day in the given range to the AmountWithDeviation
        /// structure representing the expected account balance and its standard deviation.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecast parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public virtual void ComputeDayForecast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountWithDeviation> dayForecast)
        {
            firstDay = firstDay.Date;
            lastDay = lastDay.Date;
            if (firstDay >= lastDay)
            {
                throw new ArgumentException("First day must be strictly before the last day");
            }
            DateTime date = firstDay;
            while (date <= lastDay)
            {
                if (!dayForecast.ContainsKey(date))
                {
                    dayForecast[date] = AmountWithDeviation.Zero;
                }
                dayForecast[date] += this.ExpectedBalance(date);
                date = date.AddTicks(TimeSpan.TicksPerDay);
            }
        }

        /// <summary>
        /// Computes the month forecast, mapping every month in the given range to the AmountWithDeviation 
        /// structure representing the expected account balance and its standard deviation.
        /// </summary>
        /// <param name="firstMonth">First month of the month forecast.</param>
        /// <param name="lastMonth">Last month of the month forecast.</param>
        /// <param name="monthForecast">
        /// Month forecast to be computed, mapping every day in the given range to the AmountWithDeviation
        /// structure representing the expected account balance and its standard deviation.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the monthForecast parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public virtual void ComputeMonthForecast(DateTime firstMonth, DateTime lastMonth, SortedList<DateTime, AmountWithDeviation> monthForecast)
        {
            firstMonth = new DateTime(firstMonth.Year, firstMonth.Month, 1);
            lastMonth = new DateTime(lastMonth.Year, lastMonth.Month, 1);
            if (firstMonth >= lastMonth)
            {
                throw new ArgumentException("First month must be strictly before the last month");
            }
            DateTime firstDay = firstMonth;
            DateTime lastDay = lastMonth.AddMonths(1).AddTicks(-TimeSpan.TicksPerDay);
            SortedList<DateTime, AmountWithDeviation> dayForecast = GetZeroDayForecast(firstDay, lastDay);
            this.ComputeDayForecast(firstDay, lastDay, dayForecast);
            DateTime month = firstMonth;
            while (month <= lastMonth)
            {
                if (!monthForecast.ContainsKey(month))
                {
                    monthForecast[month] = AmountWithDeviation.Zero;
                }
                monthForecast[month] += dayForecast[month.AddMonths(1).AddTicks(-TimeSpan.TicksPerDay)];
                month = month.AddMonths(1);
            }
        }

        /// <summary>
        /// Computes the explanation of the day forecast, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast explanation.</param>
        /// <param name="lastDay">Last day of the day forecast explanation.</param>
        /// <param name="dayForecastExplanation">
        /// Explanation of the day forecast to be computed, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecastExplanation parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public virtual void ExplainDayForcast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountExplanation> dayForecastExplanation)
        {
            firstDay = firstDay.Date;
            lastDay = lastDay.Date;
            if (firstDay >= lastDay)
            {
                throw new ArgumentException("First day must be strictly before the last day");
            }
            DateTime date = firstDay;
            while (date <= lastDay)
            {
                if (!dayForecastExplanation.ContainsKey(date))
                {
                    dayForecastExplanation[date] = AmountExplanation.Zero;
                }
                dayForecastExplanation[date] += this.Explain(date);
                date = date.AddTicks(TimeSpan.TicksPerDay);
            }
        }

        /// <summary>
        /// Computes the explanation of the month forecast, mapping every month in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified month.
        /// </summary>
        /// <param name="firstMonth">First month of the month forecast explanation.</param>
        /// <param name="lastMonth">Last month of the month forecast explanation.</param>
        /// <param name="monthForecastExplanation">
        /// Explanation of the month forecast to be computed, mapping every month in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified month.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the monthForecastExplanation parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public virtual void ExplainMonthForcast(DateTime firstMonth, DateTime lastMonth, SortedList<DateTime, AmountExplanation> monthForecastExplanation)
        {
            firstMonth = new DateTime(firstMonth.Year, firstMonth.Month, 1);
            lastMonth = new DateTime(lastMonth.Year, lastMonth.Month, 1);
            if (firstMonth >= lastMonth)
            {
                throw new ArgumentException("First month must be strictly before the last month");
            }
            DateTime firstDay = firstMonth;
            DateTime lastDay = lastMonth.AddMonths(1).AddTicks(-TimeSpan.TicksPerDay);
            SortedList<DateTime, AmountExplanation> dayForecastExplanation = GetZeroDayExplanation(firstDay, lastDay);
            this.ExplainDayForcast(firstDay, lastDay, dayForecastExplanation);
            foreach (KeyValuePair<DateTime, AmountExplanation> dayPair in dayForecastExplanation)
            {
                DateTime month = new DateTime(dayPair.Key.Year, dayPair.Key.Month, 1);
                if (!monthForecastExplanation.ContainsKey(month))
                {
                    monthForecastExplanation[month] = AmountExplanation.Zero;
                }
                monthForecastExplanation[month] += dayPair.Value;
            }
        }

        /// <summary>
        /// Computes the explanation of the particular period.
        /// </summary>
        /// <param name="firstDay">First day of the period.</param>
        /// <param name="lastDay">Last day of the period.</param>
        /// <returns>AmountExplanation structure representing the explanation of the particular period.</returns>
        public virtual AmountExplanation ExplainPeriod(DateTime firstDay, DateTime lastDay)
        {
            SortedList<DateTime, AmountExplanation> dayForecastExplanation = GetZeroDayExplanation(firstDay, lastDay);
            this.ExplainDayForcast(firstDay, lastDay, dayForecastExplanation);
            AmountExplanation periodExplanation = AmountExplanation.Zero;
            foreach (KeyValuePair<DateTime, AmountExplanation> explanationPair in dayForecastExplanation)
            {
                periodExplanation += explanationPair.Value;
            }
            return periodExplanation;
        }

        /// <summary>
        /// Set a new start date for the account.
        /// </summary>
        /// <param name="date">Start date of the account.</param>
        public abstract void AdjustStartDate(DateTime date);

        /// <summary>
        /// Returns the constant account balance of this account in the case when the account balance 
        /// of this account does not change from the given specified date. Otherwise, returns null.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether this account is constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the constant account balance of this account 
        /// in the case when the account balance of this account does not change from the given specified date. 
        /// Otherwise, null.
        /// </returns>
        public abstract AmountWithDeviation? IsConstant(DateTime fromDate);

        /// <summary>
        /// Clears all the constant subaccounts of this account, that are constant from the given specified date, 
        /// and returns the total sum of their balances.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether the subaccounts are constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the total sum of balances of all constant subaccounts 
        /// of this account, which are constant from the given specified date.
        /// </returns>
        public abstract AmountWithDeviation ClearConstantSubAccounts(DateTime fromDate);

        /// <summary>
        /// Sorts the account.
        /// </summary>
        public virtual void Sort()
        {

        }

        /// <summary>
        /// Compares the current object with another object of the same type.
        /// </summary>
        /// <param name="other">An object to compare with this object.</param>
        /// <returns>
        /// A 32-bit signed integer that indicates the relative order of the objects
        /// being compared. The return value has the following meanings: Value Meaning
        /// Less than zero This object is less than the other parameter. Zero This object
        /// is equal to other. Greater than zero This object is greater than other.
        /// </returns>
        public virtual int CompareTo(Account other)
        {
            return this.ToString().CompareTo(other.ToString());
        }

        /// <summary>
        /// Returns the title of the account.
        /// </summary>
        /// <returns>The string representation of this instance.</returns>
        public override string ToString()
        {
            return this.Title;
        }
    }
}
