﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CashCalc.Base
{
    /// <summary>
    /// Structure representing the explanation of the total change of the account balance at the specified date.
    /// </summary>
    [Serializable]
    public struct AmountExplanation
    {
        /// <summary>
        /// Total change of the account balance and its standard deviation at the specified date.
        /// </summary>
        public AmountWithDeviation Amount { get; set; }

        /// <summary>
        /// The explanation of the total change of the account balance and its standard deviation at the specified date.
        /// </summary>
        public List<Tuple<Account, AmountWithDeviation>> Explanation { get; set; }

        /// <summary>
        /// Empty explanation of the zero total change and zero standard deviation.
        /// </summary>
        public static readonly AmountExplanation Zero = new AmountExplanation
        {
            Amount = AmountWithDeviation.Zero,
            Explanation = new List<Tuple<Account, AmountWithDeviation>>()
        };

        /// <summary>
        /// Overloaded operator +.
        /// </summary>
        /// <param name="lhs">Left-hand side.</param>
        /// <param name="rhs">Right-hand side.</param>
        /// <returns>
        /// Returns the instance of the AmountExplanation structure representing the addition (lhs + rhs) 
        /// of the two other instances of the AmountExplanation structure.
        /// </returns>
        public static AmountExplanation operator +(AmountExplanation lhs, AmountExplanation rhs)
        {
            return Add(lhs, rhs, +1);
        }

        /// <summary>
        /// Overloaded operator -.
        /// </summary>
        /// <param name="lhs">Left-hand side.</param>
        /// <param name="rhs">Right-hand side.</param>
        /// <returns>
        /// Returns the instance of the AmountExplanation structure representing the substraction (lhs - rhs) 
        /// of the two other instances of the AmountExplanation structure.
        /// </returns>
        public static AmountExplanation operator -(AmountExplanation lhs, AmountExplanation rhs)
        {
            return Add(lhs, rhs, -1);
        }

        /// <summary>
        /// Returns the sum of the left-hand side and the right-hand side multiplied by the sign.
        /// </summary>
        /// <param name="lhs">Left-hand side.</param>
        /// <param name="rhs">Right-hand side.</param>
        /// <param name="sign">Sign.</param>
        /// <returns>
        /// Returns the instance of the AmountExplanation structure representing the addition 
        /// (lhs + sign * rhs) of the two other instances of the AmountExplanation structure.
        /// </returns>
        private static AmountExplanation Add(AmountExplanation lhs, AmountExplanation rhs, int sign)
        {
            int maxCapacity = Math.Max(1, lhs.Explanation.Count + rhs.Explanation.Count);
            List<Tuple<Account, AmountWithDeviation>> unionList = new List<Tuple<Account, AmountWithDeviation>>(maxCapacity);
            Dictionary<Account, int> accountMap = new Dictionary<Account, int>(maxCapacity);
            foreach (Tuple<Account, AmountWithDeviation> explanation in lhs.Explanation)
            {
                AddExplanation(explanation, +1, unionList, accountMap);
            }
            foreach (Tuple<Account, AmountWithDeviation> explanation in rhs.Explanation)
            {
                AddExplanation(explanation, sign, unionList, accountMap);
            }
            return new AmountExplanation
            {
                Amount = lhs.Amount + (rhs.Amount * sign),
                Explanation = unionList
            };
        }

        /// <summary>
        /// Adds the explanation to the auxiliary structures unionList and accountMap.
        /// </summary>
        /// <param name="explanation">Tuple representing the explanation.</param>
        /// <param name="sign">Sign.</param>
        /// <param name="unionList">List of explanations.</param>
        /// <param name="accountMap">Map of the accounts to the corresponding indices in the list of explanations.</param>
        private static void AddExplanation(
            Tuple<Account, AmountWithDeviation> explanation, int sign,
            List<Tuple<Account, AmountWithDeviation>> unionList, 
            Dictionary<Account, int> accountMap)
        {
            if (accountMap.ContainsKey(explanation.Item1))
            {
                int index = accountMap[explanation.Item1];
                unionList[index] = Tuple.Create(explanation.Item1, unionList[index].Item2 + (explanation.Item2 * sign));
            }
            else
            {
                unionList.Add(Tuple.Create(explanation.Item1, explanation.Item2 * sign));
                accountMap[explanation.Item1] = unionList.Count - 1;
            }
        }
    }
}
