﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;

namespace CashCalc.Base
{
    /// <summary>
    /// Structure representing the amount with its standard deviation.
    /// </summary>
    [Serializable]
    [XmlRoot("AmountWithDeviation")]
    public struct AmountWithDeviation
    {
        /// <summary>
        /// Amount.
        /// </summary>
        public decimal Amount { get; set; }

        /// <summary>
        /// Standard deviation of the amount.
        /// </summary>
        public double AmountDeviation { get; set; }

        /// <summary>
        /// Zero amount with zero standard deviation.
        /// </summary>
        public static readonly AmountWithDeviation Zero = new AmountWithDeviation { Amount = 0, AmountDeviation = 0.0 };

        /// <summary>
        /// Overloaded operator +.
        /// </summary>
        /// <param name="lhs">Left-hand side.</param>
        /// <param name="rhs">Right-hand side.</param>
        /// <returns>
        /// Returns the instance of the AmountWithDeviation structure representing the addition (lhs + rhs) 
        /// of the two other instances of the AmountWithDeviation structure.
        /// </returns>
        public static AmountWithDeviation operator +(AmountWithDeviation lhs, AmountWithDeviation rhs)
        {
            return new AmountWithDeviation
            {
                Amount = lhs.Amount + rhs.Amount,
                AmountDeviation = Math.Sqrt(lhs.AmountDeviation * lhs.AmountDeviation + rhs.AmountDeviation * rhs.AmountDeviation)
            };
        }

        /// <summary>
        /// Overloaded operator +.
        /// </summary>
        /// <param name="lhs">Left-hand side.</param>
        /// <param name="rhs">Right-hand side.</param>
        /// <returns>
        /// Returns the instance of the AmountWithDeviation structure representing the addition (lhs + rhs) 
        /// of the instance of the AmountWithDeviation structure and the decimal amount.
        /// </returns>
        public static AmountWithDeviation operator +(AmountWithDeviation lhs, decimal rhs)
        {
            return new AmountWithDeviation
            {
                Amount = lhs.Amount + rhs,
                AmountDeviation = lhs.AmountDeviation
            };
        }

        /// <summary>
        /// Overloaded operator -.
        /// </summary>
        /// <param name="lhs">Left-hand side.</param>
        /// <param name="rhs">Right-hand side.</param>
        /// <returns>
        /// Returns the instance of the AmountWithDeviation structure representing the substraction (lhs - rhs) 
        /// of the two other instances of the AmountWithDeviation structure.
        /// </returns>
        public static AmountWithDeviation operator -(AmountWithDeviation lhs, AmountWithDeviation rhs)
        {
            return new AmountWithDeviation
            {
                Amount = lhs.Amount - rhs.Amount,
                AmountDeviation = Math.Sqrt(lhs.AmountDeviation * lhs.AmountDeviation + rhs.AmountDeviation * rhs.AmountDeviation)
            };
        }

        /// <summary>
        /// Overloaded operator -.
        /// </summary>
        /// <param name="lhs">Left-hand side.</param>
        /// <param name="rhs">Right-hand side.</param>
        /// <returns>
        /// Returns the instance of the AmountWithDeviation structure representing the substraction (lhs - rhs) 
        /// of the instance of the AmountWithDeviation structure and the decimal amount.
        /// </returns>
        public static AmountWithDeviation operator -(AmountWithDeviation lhs, decimal rhs)
        {
            return new AmountWithDeviation
            {
                Amount = lhs.Amount - rhs,
                AmountDeviation = lhs.AmountDeviation
            };
        }

        /// <summary>
        /// Overloaded operator *.
        /// </summary>
        /// <param name="lhs">Left-hand side.</param>
        /// <param name="rhs">Right-hand side.</param>
        /// <returns>
        /// Returns the instance of the AmountWithDeviation structure representing the multiplication (lhs * rhs)
        /// of the instance of the AmountWithDeviation structure and the integer amount.
        /// </returns>
        public static AmountWithDeviation operator *(AmountWithDeviation lhs, int rhs)
        {
            return new AmountWithDeviation
            {
                Amount = rhs * lhs.Amount,
                AmountDeviation = Math.Sqrt(Math.Abs(rhs)) * lhs.AmountDeviation
            };
        }

        /// <summary>
        /// Returns the string representation of the AmountWithDeviation structure.
        /// </summary>
        /// <returns>The string representation of this instance.</returns>
        public override string ToString()
        {
            return String.Format("{0} +- {1}", this.Amount, this.AmountDeviation);
        }
    }
}
