﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;

namespace CashCalc.Base
{
    /// <summary>
    /// Account with a CheckBox.
    /// </summary>
    [Serializable]
    [XmlRoot("CheckBoxAccount")]
    public sealed class CheckBoxAccount : Account
    {
        /// <summary>
        /// Default constructor of the CheckBoxAccount class (due to the XML serialization).
        /// </summary>
        public CheckBoxAccount()
        {

        }

        /// <summary>
        /// Constructor of the CheckBoxAccount class. The account is enabled by default.
        /// </summary>
        /// <param name="account">Account.</param>
        public CheckBoxAccount(Account account)
            : this(account, true)
        {

        }

        /// <summary>
        /// Constructor of the CheckBoxAccount structure.
        /// </summary>
        /// <param name="account">Account.</param>
        /// <param name="enabled">CheckBox indicator whether the account is enabled or not.</param>
        public CheckBoxAccount(Account account, bool enabled)
        {            
            this.Account = account;
            this.Enabled = enabled;
        }

        /// <summary>
        /// Account title.
        /// </summary>
        [XmlIgnore]
        public override string Title
        {
            get { if (this.Account != null) return this.Account.Title; else return null; }
            set { if (this.Account != null) this.Account.Title = value; }
        }

        /// <summary>
        /// Account.
        /// </summary>
        [XmlElement(Type = typeof(CompositeAccount)),
        XmlElement(Type = typeof(ConstantAccount)),
        XmlElement(Type = typeof(ContinuousMonthAccount)),
        XmlElement(Type = typeof(EveryNthDayAccount)),
        XmlElement(Type = typeof(EveryNthMonthAccount)),
        XmlElement(Type = typeof(WeekAccount))]
        public Account Account { get; set; }

        /// <summary>
        /// CheckBox indicator whether the account is enabled or not.
        /// </summary>
        public bool Enabled { get; set; }

        /// <summary>
        /// Computes the expected account balance and its standard deviation at the specified date.
        /// </summary>
        /// <param name="date">Date of the forecast.</param>
        /// <returns>
        /// Returns the AmountWithDeviation structure representing the expected account balance 
        /// and its standard deviation at the specified date.
        /// </returns>
        public override AmountWithDeviation ExpectedBalance(DateTime date)
        {
            if (this.Enabled)
            {
                return this.Account.ExpectedBalance(date);
            }
            else
            {
                return AmountWithDeviation.Zero;
            }
        }

        /// <summary>
        /// Returns the explanation of the total change of the account balance at the specified date.
        /// </summary>
        /// <param name="date">Date of the account balance change.</param>
        /// <returns>
        /// Returns the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </returns>
        public override AmountExplanation Explain(DateTime date)
        {
            if (this.Enabled)
            {
                return this.Account.Explain(date);
            }
            else
            {
                return AmountExplanation.Zero;
            }
        }

        /// <summary>
        /// Computes the day forecast, mapping every day in the given range to the AmountWithDeviation 
        /// structure representing the expected account balance and its standard deviation.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast.</param>
        /// <param name="lastDay">Last day of the day forecast.</param>
        /// <param name="dayForecast">
        /// Day forecast to be computed, mapping every day in the given range to the AmountWithDeviation
        /// structure representing the expected account balance and its standard deviation.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecast parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public override void ComputeDayForecast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountWithDeviation> dayForecast)
        {
            if (this.Enabled)
            {
                this.Account.ComputeDayForecast(firstDay, lastDay, dayForecast);
            }
        }

        /// <summary>
        /// Computes the explanation of the day forecast, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast explanation.</param>
        /// <param name="lastDay">Last day of the day forecast explanation.</param>
        /// <param name="dayForecastExplanation">
        /// Explanation of the day forecast to be computed, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecastExplanation parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public override void ExplainDayForcast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountExplanation> dayForecastExplanation)
        {
            if (this.Enabled)
            {
                this.Account.ExplainDayForcast(firstDay, lastDay, dayForecastExplanation);
            }
        }

        /// <summary>
        /// Set a new start date for the account.
        /// </summary>
        /// <param name="date">Start date of the account.</param>
        public override void AdjustStartDate(DateTime date)
        {
            this.Account.AdjustStartDate(date);
        }

        /// <summary>
        /// Returns the constant account balance of this account in the case when the account balance 
        /// of this account does not change from the given specified date. Otherwise, returns null.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether this account is constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the constant account balance of this account 
        /// in the case when the account balance of this account does not change from the given specified date. 
        /// Otherwise, null.
        /// </returns>
        public override AmountWithDeviation? IsConstant(DateTime fromDate)
        {
            return this.Account.IsConstant(fromDate);
        }

        /// <summary>
        /// Clears all the constant subaccounts of this account, that are constant from the given specified date, 
        /// and returns the total sum of their balances.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether the subaccounts are constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the total sum of balances of all constant subaccounts 
        /// of this account, which are constant from the given specified date.
        /// </returns>
        public override AmountWithDeviation ClearConstantSubAccounts(DateTime fromDate)
        {
            return this.Account.ClearConstantSubAccounts(fromDate);
        }

        /// <summary>
        /// Sorts the account.
        /// </summary>
        public override void Sort()
        {
            this.Account.Sort();
        }

        /// <summary>
        /// Returns the title of the account.
        /// </summary>
        /// <returns>The string representation of this instance.</returns>
        public override string ToString()
        {
            return this.Account.ToString();
        }
    }
}
