﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;

namespace CashCalc.Base
{
    /// <summary>
    /// Account which consists of several accounts.
    /// </summary>
    [Serializable]
    [XmlRoot("CompositeAccount")]
    public sealed class CompositeAccount : Account
    {
        /// <summary>
        /// Default constructor of the CompositeAccount class (due to the XML serialization).
        /// </summary>
        public CompositeAccount()
        {
            this.Title = "Composite Account";
            this.CheckBoxAccounts = new List<CheckBoxAccount>();
        }

        /// <summary>
        /// Constructor of the CompositeAccount class.
        /// </summary>
        /// <param name="title">Title of the composite account.</param>
        public CompositeAccount(string title)
        {
            this.Title = title;
            this.CheckBoxAccounts = new List<CheckBoxAccount>();
        }

        /// <summary>
        /// List of accounts.
        /// </summary>
        public List<CheckBoxAccount> CheckBoxAccounts { get; set; }

        /// <summary>
        /// Computes the expected account balance and its standard deviation at the specified date.
        /// </summary>
        /// <param name="date">Date of the forecast.</param>
        /// <returns>
        /// Returns the AmountWithDeviation structure representing the expected account balance 
        /// and its standard deviation at the specified date.
        /// </returns>
        public override AmountWithDeviation ExpectedBalance(DateTime date)
        {
            AmountWithDeviation expectedBalance = AmountWithDeviation.Zero;
            if (this.CheckBoxAccounts != null && this.CheckBoxAccounts.Count > 0)
            {
                foreach (CheckBoxAccount checkBoxAccount in this.CheckBoxAccounts.Where(account => account.Enabled))
                {
                    expectedBalance += checkBoxAccount.ExpectedBalance(date);
                }
            }
            return expectedBalance;
        }

        /// <summary>
        /// Returns the explanation of the total change of the account balance at the specified date.
        /// </summary>
        /// <param name="date">Date of the account balance change.</param>
        /// <returns>
        /// Returns the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </returns>
        public override AmountExplanation Explain(DateTime date)
        {
            AmountExplanation explanation = AmountExplanation.Zero;
            if (this.CheckBoxAccounts != null && this.CheckBoxAccounts.Count > 0)
            {
                foreach (CheckBoxAccount checkBoxAccount in this.CheckBoxAccounts.Where(account => account.Enabled))
                {
                    explanation += checkBoxAccount.Explain(date);
                }
            }
            return explanation;
        }

        /// <summary>
        /// Computes the day forecast, mapping every day in the given range to the AmountWithDeviation 
        /// structure representing the expected account balance and its standard deviation.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast.</param>
        /// <param name="lastDay">Last day of the day forecast.</param>
        /// <param name="dayForecast">
        /// Day forecast to be computed, mapping every day in the given range to the AmountWithDeviation
        /// structure representing the expected account balance and its standard deviation.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecast parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public override void ComputeDayForecast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountWithDeviation> dayForecast)
        {
            if (this.CheckBoxAccounts != null && this.CheckBoxAccounts.Count > 0)
            {
                foreach (CheckBoxAccount checkBoxAccount in this.CheckBoxAccounts.Where(account => account.Enabled))
                {
                    checkBoxAccount.ComputeDayForecast(firstDay, lastDay, dayForecast);
                }
            }
        }

        /// <summary>
        /// Computes the explanation of the day forecast, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast explanation.</param>
        /// <param name="lastDay">Last day of the day forecast explanation.</param>
        /// <param name="dayForecastExplanation">
        /// Explanation of the day forecast to be computed, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecastExplanation parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public override void ExplainDayForcast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountExplanation> dayForecastExplanation)
        {
            if (this.CheckBoxAccounts != null && this.CheckBoxAccounts.Count > 0)
            {
                foreach (CheckBoxAccount checkBoxAccount in this.CheckBoxAccounts.Where(account => account.Enabled))
                {
                    checkBoxAccount.ExplainDayForcast(firstDay, lastDay, dayForecastExplanation);
                }
            }
        }

        /// <summary>
        /// Set a new start date for the account.
        /// </summary>
        /// <param name="date">Start date of the account.</param>
        public override void AdjustStartDate(DateTime date)
        {
            if (this.CheckBoxAccounts != null && this.CheckBoxAccounts.Count > 0)
            {
                foreach (CheckBoxAccount checkBoxAccount in this.CheckBoxAccounts)
                {
                    checkBoxAccount.AdjustStartDate(date);
                }
            }
        }

        /// <summary>
        /// Returns the constant account balance of this account in the case when the account balance 
        /// of this account does not change from the given specified date. Otherwise, returns null.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether this account is constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the constant account balance of this account 
        /// in the case when the account balance of this account does not change from the given specified date. 
        /// Otherwise, null.
        /// </returns>
        public override AmountWithDeviation? IsConstant(DateTime fromDate)
        {
            return null;
        }

        /// <summary>
        /// Clears all the constant subaccounts of this account, that are constant from the given specified date, 
        /// and returns the total sum of their balances.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether the subaccounts are constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the total sum of balances of all constant subaccounts 
        /// of this account, which are constant from the given specified date.
        /// </returns>
        public override AmountWithDeviation ClearConstantSubAccounts(DateTime fromDate)
        {
            fromDate = fromDate.Date;
            AmountWithDeviation totalConstantAmount = AmountWithDeviation.Zero;
            List<CheckBoxAccount> newCheckBoxAccounts = new List<CheckBoxAccount>(this.CheckBoxAccounts.Count);
            foreach (CheckBoxAccount checkBoxAccount in this.CheckBoxAccounts)
            {
                AmountWithDeviation? constant = checkBoxAccount.IsConstant(fromDate);
                if (constant.HasValue)
                {
                    totalConstantAmount += constant.Value;
                }
                else
                {
                    totalConstantAmount += checkBoxAccount.ClearConstantSubAccounts(fromDate);
                    newCheckBoxAccounts.Add(checkBoxAccount);
                }
            }
            this.CheckBoxAccounts = newCheckBoxAccounts;
            return totalConstantAmount;
        }

        /// <summary>
        /// Sorts the account.
        /// </summary>
        public override void Sort()
        {
            this.CheckBoxAccounts.Sort();
        }

        /// <summary>
        /// Returns the title of the composite account.
        /// </summary>
        /// <returns>The string representation of this instance.</returns>
        public override string ToString()
        {
            return "[" + this.Title + "]";
        }
    }
}
