﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;

namespace CashCalc.Base
{
    /// <summary>
    /// Account with a constant account balance from the specified date.
    /// </summary>
    [Serializable]
    [XmlRoot("ConstantAccount")]
    public sealed class ConstantAccount : Account
    {
        /// <summary>
        /// Default constructor of the ConstantAccount class (due to the XML serialization).
        /// </summary>
        public ConstantAccount()
            : this(System.DateTime.Today, 0, 0.0)
        {

        }

        /// <summary>
        /// Constructor of the ConstantAccount class.
        /// </summary>
        /// <param name="startDate">First date when the account achieves its specified account balance.</param>
        /// <param name="amount">Account balance achieved at the specified date.</param>
        /// <param name="amountDeviation">Standard deviation of the account balance achieved at the specified date.</param>
        public ConstantAccount(DateTime startDate, decimal amount, double amountDeviation)
        {
            this.Title = "Constant Account";
            this.StartDate = startDate.Date;
            this.Amount = new AmountWithDeviation
            {
                Amount = amount,
                AmountDeviation = amountDeviation
            };
        }

        /// <summary>
        /// First date when the account achieves its specified account balance.
        /// </summary>
        public DateTime StartDate { get; set; }

        /// <summary>
        /// Account balance and its standard deviation achieved at the specified date.
        /// </summary>
        public AmountWithDeviation Amount { get; set; }

        /// <summary>
        /// Computes the expected account balance and its standard deviation at the specified date.
        /// </summary>
        /// <param name="date">Date of the forecast.</param>
        /// <returns>
        /// Returns the AmountWithDeviation structure representing the expected account balance 
        /// and its standard deviation at the specified date.
        /// </returns>
        public override AmountWithDeviation ExpectedBalance(DateTime date)
        {
            if (date.Date >= this.StartDate)
            {
                return this.Amount;
            }
            else
            {
                return AmountWithDeviation.Zero;
            }
        }

        /// <summary>
        /// Returns the explanation of the total change of the account balance at the specified date.
        /// </summary>
        /// <param name="date">Date of the account balance change.</param>
        /// <returns>
        /// Returns the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </returns>
        public override AmountExplanation Explain(DateTime date)
        {
            if (date.Date == this.StartDate)
            {
                return new AmountExplanation
                {
                    Amount = this.Amount,
                    Explanation = new List<Tuple<Account, AmountWithDeviation>>
                    {
                        Tuple.Create(this as Account, this.Amount)
                    }
                };
            }
            else
            {
                return AmountExplanation.Zero;
            }
        }

        /// <summary>
        /// Set a new start date for the account.
        /// </summary>
        /// <param name="date">Start date of the account.</param>
        public override void AdjustStartDate(DateTime date)
        {
            if (date.Date > this.StartDate)
            {
                this.Amount = new AmountWithDeviation { Amount = this.Amount.Amount };
            }
        }

        /// <summary>
        /// Returns the constant account balance of this account in the case when the account balance 
        /// of this account does not change from the given specified date. Otherwise, returns null.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether this account is constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the constant account balance of this account 
        /// in the case when the account balance of this account does not change from the given specified date. 
        /// Otherwise, null.
        /// </returns>
        public override AmountWithDeviation? IsConstant(DateTime fromDate)
        {
            if (fromDate.Date >= this.StartDate)
            {
                return this.Amount;
            }
            else
            {
                return null;
            }
        }

        /// <summary>
        /// Clears all the constant subaccounts of this account, that are constant from the given specified date, 
        /// and returns the total sum of their balances.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether the subaccounts are constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the total sum of balances of all constant subaccounts 
        /// of this account, which are constant from the given specified date.
        /// </returns>
        public override AmountWithDeviation ClearConstantSubAccounts(DateTime fromDate)
        {
            return AmountWithDeviation.Zero;
        }
    }
}
