﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;

namespace CashCalc.Base
{
    /// <summary>
    /// Account whose balance changes every day by a constant amount.
    /// </summary>
    [Serializable]
    [XmlRoot("ContinuousMonthAccount")]
    public sealed class ContinuousMonthAccount : Account
    {
        /// <summary>
        /// Default constructor of the ContinuousMonthAccount class (due to the XML serialization).
        /// </summary>
        public ContinuousMonthAccount()
            : this(System.DateTime.Today, System.DateTime.Today.AddYears(1), 0, 0.0)
        {

        }

        /// <summary>
        /// Constructor of the ContinuousMonthAccount class.
        /// </summary>
        /// <param name="startDate">First date when the account changes.</param>
        /// <param name="endDate">Last date when the account changes.</param>
        /// <param name="amount30">Amount of the 30 day change of the account balance.</param>
        /// <param name="amount30Deviation">Standard deviation of the 30 day change of the account balance.</param>
        public ContinuousMonthAccount(DateTime startDate, DateTime endDate, decimal amount30, double amount30Deviation)
        {
            this.Title = "Continuous Month Account";
            this.StartDate = startDate.Date;
            this.EndDate = endDate.Date;
            this.Amount30 = new AmountWithDeviation
            {
                Amount = amount30,
                AmountDeviation = amount30Deviation
            };
        }

        /// <summary>
        /// First date when the account changes.
        /// </summary>
        public DateTime StartDate { get; set; }

        /// <summary>
        /// Last date when the account changes.
        /// </summary>
        public DateTime EndDate { get; set; }

        /// <summary>
        /// Amount of the 30 day change and its standard deviation.
        /// </summary>
        public AmountWithDeviation Amount30 { get; set; }

        /// <summary>
        /// Computes the expected account balance and its standard deviation at the specified date.
        /// </summary>
        /// <param name="date">Date of the forecast.</param>
        /// <returns>
        /// Returns the AmountWithDeviation structure representing the expected account balance 
        /// and its standard deviation at the specified date.
        /// </returns>
        public override AmountWithDeviation ExpectedBalance(DateTime date)
        {
            date = date.Date;
            if (date >= this.StartDate)
            {
                if (date >= this.EndDate)
                {
                    date = this.EndDate;
                }
                return new AmountWithDeviation
                {
                    Amount = (decimal)((date - this.StartDate).TotalDays * (double)this.Amount30.Amount / 30.0),
                    AmountDeviation = Math.Sqrt((date - this.StartDate).TotalDays) * this.Amount30.AmountDeviation / Math.Sqrt(30.0)
                };
            }
            else
            {
                return AmountWithDeviation.Zero;
            }
        }

        /// <summary>
        /// Returns the explanation of the total change of the account balance at the specified date.
        /// </summary>
        /// <param name="date">Date of the account balance change.</param>
        /// <returns>
        /// Returns the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </returns>
        public override AmountExplanation Explain(DateTime date)
        {
            date = date.Date;
            if (date >= this.StartDate && date <= this.EndDate)
            {
                AmountWithDeviation oneDayAmount = new AmountWithDeviation
                {
                    Amount = this.Amount30.Amount / 30.0m,
                    AmountDeviation = this.Amount30.AmountDeviation / Math.Sqrt(30.0)
                };
                return new AmountExplanation
                {
                    Amount = oneDayAmount,
                    Explanation = new List<Tuple<Account, AmountWithDeviation>>
                    {
                        Tuple.Create(this as Account, oneDayAmount)
                    }
                };
            }
            else
            {
                return AmountExplanation.Zero;
            }
        }

        /// <summary>
        /// Set a new start date for the account.
        /// </summary>
        /// <param name="date">Start date of the account.</param>
        public override void AdjustStartDate(DateTime date)
        {
            if (date.Date > this.StartDate.Date)
            {
                this.StartDate = date.Date;
            }
        }

        /// <summary>
        /// Returns the constant account balance of this account in the case when the account balance 
        /// of this account does not change from the given specified date. Otherwise, returns null.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether this account is constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the constant account balance of this account 
        /// in the case when the account balance of this account does not change from the given specified date. 
        /// Otherwise, null.
        /// </returns>
        public override AmountWithDeviation? IsConstant(DateTime fromDate)
        {
            fromDate = fromDate.Date;
            if (fromDate >= this.EndDate)
            {
                return this.ExpectedBalance(this.EndDate);
            }
            else
            {
                return null;
            }
        }

        /// <summary>
        /// Clears all the constant subaccounts of this account, that are constant from the given specified date, 
        /// and returns the total sum of their balances.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether the subaccounts are constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the total sum of balances of all constant subaccounts 
        /// of this account, which are constant from the given specified date.
        /// </returns>
        public override AmountWithDeviation ClearConstantSubAccounts(DateTime fromDate)
        {
            return AmountWithDeviation.Zero;
        }
    }
}
