﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CashCalc.Common;

namespace CashCalc.Base
{
    /// <summary>
    /// Base abstract class for all enumerable accounts whose balance changes at specific dates by a constant amount.
    /// </summary>
    [Serializable]
    public abstract class EnumerableAccount : Account
    {
        /// <summary>
        /// First date when the account can (possibly) change.
        /// </summary>
        public DateTime StartDate { get; set; }

        /// <summary>
        /// Last date when the account can (possibly) change.
        /// </summary>
        public DateTime EndDate { get; set; }

        /// <summary>
        /// Enumeration of all dates when the account balance changes by a constant amount.
        /// </summary>
        public abstract IEnumerable<DateTime> Dates { get; }

        /// <summary>
        /// Amount of change of the account balance and its standard deviation.
        /// </summary>
        public AmountWithDeviation Amount { get; set; }

        /// <summary>
        /// Exceptions.
        /// </summary>
        public SerializableSortedList<DateTime, AmountWithDeviation> Exceptions { get; set; }

        /// <summary>
        /// Enumeration of all dates of the account including the exceptions.
        /// </summary>
        public IEnumerable<DateTime> DatesWithExceptions
        {
            get
            {
                IEnumerator<DateTime> datesEnumerator = this.Dates.GetEnumerator();
                IEnumerator<DateTime> exceptionsEnumerator = this.Exceptions.Keys.GetEnumerator();
                bool validCurrentDate = datesEnumerator.MoveNext();
                bool validCurrentExceptionDate = exceptionsEnumerator.MoveNext();

                while (validCurrentDate && validCurrentExceptionDate)
                {
                    DateTime date = datesEnumerator.Current;
                    DateTime exceptionDate = exceptionsEnumerator.Current;
                    if (date < exceptionDate)
                    {
                        if (date >= this.StartDate && date <= this.EndDate) { yield return date; } else { yield break; }
                        validCurrentDate = datesEnumerator.MoveNext();
                    }
                    else if (date == exceptionDate)
                    {
                        if (date >= this.StartDate && date <= this.EndDate) { yield return date; } else { yield break; }
                        validCurrentDate = datesEnumerator.MoveNext();
                        validCurrentExceptionDate = exceptionsEnumerator.MoveNext();
                    }
                    else if (date > exceptionDate)
                    {
                        if (exceptionDate >= this.StartDate && exceptionDate <= this.EndDate) { yield return exceptionDate; } else { yield break; }
                        validCurrentExceptionDate = exceptionsEnumerator.MoveNext();
                    }
                }

                while (validCurrentDate)
                {
                    DateTime date = datesEnumerator.Current;
                    if (date >= this.StartDate && date <= this.EndDate) { yield return date; } else { yield break; }
                    validCurrentDate = datesEnumerator.MoveNext();
                }

                while (validCurrentExceptionDate)
                {
                    DateTime exceptionDate = exceptionsEnumerator.Current;
                    if (exceptionDate >= this.StartDate && exceptionDate <= this.EndDate) { yield return exceptionDate; } else { yield break; }
                    validCurrentExceptionDate = exceptionsEnumerator.MoveNext();
                }
            }
        }

        /// <summary>
        /// Computes the expected account balance and its standard deviation at the specified date.
        /// </summary>
        /// <param name="date">Date of the forecast.</param>
        /// <returns>
        /// Returns the AmountWithDeviation structure representing the expected account balance 
        /// and its standard deviation at the specified date.
        /// </returns>
        public override AmountWithDeviation ExpectedBalance(DateTime date)
        {
            AmountWithDeviation balance = AmountWithDeviation.Zero;
            foreach (DateTime stepDate in this.DatesWithExceptions)
            {
                if (stepDate <= date)
                {
                    if (this.Exceptions != null && this.Exceptions.ContainsKey(stepDate))
                    {
                        balance += this.Exceptions[stepDate];
                    }
                    else
                    {
                        balance += this.Amount;
                    }
                }
                else break;
            }
            return balance;
        }

        /// <summary>
        /// Returns the explanation of the total change of the account balance at the specified date.
        /// </summary>
        /// <param name="date">Date of the account balance change.</param>
        /// <returns>
        /// Returns the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </returns>
        public override AmountExplanation Explain(DateTime date)
        {
            foreach (DateTime stepDate in this.DatesWithExceptions)
            {
                if (stepDate == date)
                {
                    AmountWithDeviation amountChange = AmountWithDeviation.Zero;
                    if (this.Exceptions != null && this.Exceptions.ContainsKey(stepDate))
                    {
                        amountChange = this.Exceptions[stepDate];
                    }
                    else
                    {
                        amountChange = this.Amount;
                    }
                    return new AmountExplanation
                    {
                        Amount = amountChange,
                        Explanation = new List<Tuple<Account, AmountWithDeviation>>
                        {
                            Tuple.Create(this as Account, amountChange)
                        }
                    };
                }
                else if (stepDate > date)
                {
                    break;
                }
            }
            return AmountExplanation.Zero;
        }

        /// <summary>
        /// Computes the day forecast, mapping every day in the given range to the AmountWithDeviation 
        /// structure representing the expected account balance and its standard deviation.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast.</param>
        /// <param name="lastDay">Last day of the day forecast.</param>
        /// <param name="dayForecast">
        /// Day forecast to be computed, mapping every day in the given range to the AmountWithDeviation
        /// structure representing the expected account balance and its standard deviation.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecast parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public override void ComputeDayForecast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountWithDeviation> dayForecast)
        {
            firstDay = firstDay.Date;
            lastDay = lastDay.Date;
            if (firstDay >= lastDay)
            {
                throw new ArgumentException("First day must be strictly before the last day");
            }
            SortedList<DateTime, AmountWithDeviation> amountCache = new SortedList<DateTime, AmountWithDeviation>();
            foreach (DateTime stepDate in this.DatesWithExceptions)
            {
                if (stepDate <= lastDay)
                {
                    if (this.Exceptions != null && this.Exceptions.ContainsKey(stepDate))
                    {
                        amountCache[stepDate] = this.Exceptions[stepDate];
                    }
                    else
                    {
                        amountCache[stepDate] = this.Amount;
                    }
                }
                else break;
            }
            AmountWithDeviation balance = this.ExpectedBalance(firstDay.AddTicks(-TimeSpan.TicksPerDay));
            DateTime date = firstDay;
            while (date <= lastDay)
            {
                if (!dayForecast.ContainsKey(date))
                {
                    dayForecast[date] = AmountWithDeviation.Zero;
                }
                if (amountCache.ContainsKey(date))
                {
                    balance += amountCache[date];
                }
                dayForecast[date] += balance;
                date = date.AddTicks(TimeSpan.TicksPerDay);
            }
        }

        /// <summary>
        /// Computes the explanation of the day forecast, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast explanation.</param>
        /// <param name="lastDay">Last day of the day forecast explanation.</param>
        /// <param name="dayForecastExplanation">
        /// Explanation of the day forecast to be computed, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecastExplanation parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public override void ExplainDayForcast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountExplanation> dayForecastExplanation)
        {
            firstDay = firstDay.Date;
            lastDay = lastDay.Date;
            if (firstDay >= lastDay)
            {
                throw new ArgumentException("First day must be strictly before the last day");
            }
            SortedList<DateTime, AmountExplanation> explanationCache = new SortedList<DateTime, AmountExplanation>();
            foreach (DateTime stepDate in this.DatesWithExceptions)
            {
                if (stepDate <= lastDay)
                {
                    AmountWithDeviation amountChange = AmountWithDeviation.Zero;
                    if (this.Exceptions != null && this.Exceptions.ContainsKey(stepDate))
                    {
                        amountChange = this.Exceptions[stepDate];
                    }
                    else
                    {
                        amountChange = this.Amount;
                    }
                    explanationCache[stepDate] = new AmountExplanation
                    {
                        Amount = amountChange,
                        Explanation = new List<Tuple<Account, AmountWithDeviation>>
                        {
                            Tuple.Create(this as Account, amountChange)
                        }
                    };
                }
                else break;
            }
            DateTime date = firstDay;
            while (date <= lastDay)
            {
                if (!dayForecastExplanation.ContainsKey(date))
                {
                    dayForecastExplanation[date] = AmountExplanation.Zero;
                }
                if (explanationCache.ContainsKey(date))
                {
                    dayForecastExplanation[date] += explanationCache[date];
                }
                date = date.AddTicks(TimeSpan.TicksPerDay);
            }
        }

        /// <summary>
        /// Returns the constant account balance of this account in the case when the account balance 
        /// of this account does not change from the given specified date. Otherwise, returns null.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether this account is constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the constant account balance of this account 
        /// in the case when the account balance of this account does not change from the given specified date. 
        /// Otherwise, null.
        /// </returns>
        public override AmountWithDeviation? IsConstant(DateTime fromDate)
        {
            fromDate = fromDate.Date;
            bool isConstantCheck = true;
            foreach (DateTime date in this.DatesWithExceptions)
            {
                if (date > fromDate)
                {
                    isConstantCheck = false;
                    break;
                }
            }
            if (isConstantCheck)
            {
                return this.ExpectedBalance(fromDate);
            }
            else
            {
                return null;
            }
        }

        /// <summary>
        /// Clears all the constant subaccounts of this account, that are constant from the given specified date, 
        /// and returns the total sum of their balances.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether the subaccounts are constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the total sum of balances of all constant subaccounts 
        /// of this account, which are constant from the given specified date.
        /// </returns>
        public override AmountWithDeviation ClearConstantSubAccounts(DateTime fromDate)
        {
            fromDate = fromDate.Date;
            SerializableSortedList<DateTime, AmountWithDeviation> clearedExceptions =
                new SerializableSortedList<DateTime, AmountWithDeviation>();
            foreach (KeyValuePair<DateTime, AmountWithDeviation> pair in this.Exceptions)
            {
                if (pair.Key >= fromDate && pair.Key >= this.StartDate)
                {
                    clearedExceptions.Add(pair.Key, pair.Value);
                }
            }
            if (clearedExceptions.Count < this.Exceptions.Count)
            {
                AmountWithDeviation oldBalance = this.ExpectedBalance(fromDate);
                this.Exceptions = clearedExceptions;
                AmountWithDeviation newBalance = this.ExpectedBalance(fromDate);
                return oldBalance - newBalance;
            }
            else
            {
                return AmountWithDeviation.Zero;
            }
        }
    }
}
