﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;
using CashCalc.Common;

namespace CashCalc.Base
{
    /// <summary>
    /// Account whose balance changes every month by a constant amount.
    /// </summary>
    [Serializable]
    [XmlRoot("EveryNthMonthAccount")]
    public sealed class EveryNthMonthAccount : EnumerableAccount
    {
        /// <summary>
        /// Default constructor of the EveryNthMonthAccount class (due to the XML serialization).
        /// </summary>
        public EveryNthMonthAccount()
            : this(System.DateTime.Today, System.DateTime.Today.AddYears(1), 0, 0.0, 1)
        {

        }

        /// <summary>
        /// Constructor of the EveryNthMonthAccount class.
        /// </summary>
        /// <param name="startDate">First date when the account changes.</param>
        /// <param name="endDate">Last date when the account can (possibly) change.</param>
        /// <param name="amount">Amount of change of the account balance.</param>
        /// <param name="amountDeviation">Standard deviation of the change.</param>
        /// <param name="step">Distance between every two consecutive account balance changes in months.</param>
        public EveryNthMonthAccount(DateTime startDate, DateTime endDate, decimal amount, double amountDeviation, int step)
            : this(startDate, endDate, amount, amountDeviation, step, null)
        {

        }

        /// <summary>
        /// Constructor of the EveryNthMonthAccount class.
        /// </summary>
        /// <param name="startDate">First date when the account changes.</param>
        /// <param name="endDate">Last date when the account can (possibly) change.</param>
        /// <param name="amount">Amount of change of the account balance.</param>
        /// <param name="amountDeviation">Standard deviation of the change.</param>
        /// <param name="step">Distance between every two consecutive account balance changes in months.</param>
        /// <param name="exceptions">Exceptions.</param>
        public EveryNthMonthAccount(DateTime startDate, DateTime endDate, decimal amount, double amountDeviation, int step, SerializableSortedList<DateTime, AmountWithDeviation> exceptions)
        {
            this.Title = "Month Account";
            this.StartDate = startDate.Date;
            this.EndDate = endDate.Date;
            this.Amount = new AmountWithDeviation
            {
                Amount = amount,
                AmountDeviation = amountDeviation
            };
            this.Step = step;
            this.Exceptions = exceptions;
        }

        /// <summary>
        /// Distance between every two consecutive account balance changes in months.
        /// </summary>
        public int Step { get; set; }

        /// <summary>
        /// Set a new start date for the account.
        /// </summary>
        /// <param name="date">Start date of the account.</param>
        public override void AdjustStartDate(DateTime date)
        {
            DateTime startDate = this.StartDate.Date;
            date = date.Date;
            if (date > startDate)
            {
                while (date > startDate)
                {
                    startDate = startDate.AddMonths(this.Step);
                }
                this.StartDate = startDate;
            }
        }

        /// <summary>
        /// Enumeration of all dates when the account balance changes by a constant amount.
        /// </summary>
        public override IEnumerable<DateTime> Dates
        {
            get
            {
                DateTime date = this.StartDate.Date;
                while (date <= this.EndDate)
                {
                    yield return date;
                    date = date.AddMonths(this.Step);
                }
            }
        }
    }
}
