﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;

namespace CashCalc.Base
{
    /// <summary>
    /// Account which consists of inflow and outflow accounts.
    /// </summary>
    [Serializable]
    [XmlRoot("InflowOutflowAccount")]
    public sealed class InflowOutflowAccount : Account
    {
        /// <summary>
        /// Default constructor of the InflowOutflowAccount class (due to the XML serialization).
        /// </summary>
        public InflowOutflowAccount()
        {
            this.Title = "Inflow Outflow Account";
            this.InitialBalance = 0;
            this.InflowAccount = new CheckBoxAccount(new CompositeAccount("Inflow"));
            this.OutflowAccount = new CheckBoxAccount(new CompositeAccount("Outflow"));
        }

        /// <summary>
        /// Initial account balance.
        /// </summary>
        public decimal InitialBalance { get; set; }
        
        /// <summary>
        /// Inflow accounts.
        /// </summary>
        public CheckBoxAccount InflowAccount { get; set; }

        /// <summary>
        /// Outflow accounts.
        /// </summary>
        public CheckBoxAccount OutflowAccount { get; set; }

        /// <summary>
        /// Computes the expected account balance and its standard deviation at the specified date.
        /// </summary>
        /// <param name="date">Date of the forecast.</param>
        /// <returns>
        /// Returns the AmountWithDeviation structure representing the expected account balance 
        /// and its standard deviation at the specified date.
        /// </returns>
        public override AmountWithDeviation ExpectedBalance(DateTime date)
        {
            return this.InflowAccount.ExpectedBalance(date) - this.OutflowAccount.ExpectedBalance(date) + this.InitialBalance;
        }

        /// <summary>
        /// Returns the explanation of the total change of the account balance at the specified date.
        /// </summary>
        /// <param name="date">Date of the account balance change.</param>
        /// <returns>
        /// Returns the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </returns>
        public override AmountExplanation Explain(DateTime date)
        {
            return this.InflowAccount.Explain(date) - this.OutflowAccount.Explain(date);
        }

        /// <summary>
        /// Computes the day forecast, mapping every day in the given range to the AmountWithDeviation 
        /// structure representing the expected account balance and its standard deviation.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast.</param>
        /// <param name="lastDay">Last day of the day forecast.</param>
        /// <param name="dayForecast">
        /// Day forecast to be computed, mapping every day in the given range to the AmountWithDeviation
        /// structure representing the expected account balance and its standard deviation.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecast parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public override void ComputeDayForecast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountWithDeviation> dayForecast)
        {
            firstDay = firstDay.Date;
            lastDay = lastDay.Date;
            SortedList<DateTime, AmountWithDeviation> inflowDayForecast = Account.GetZeroDayForecast(firstDay, lastDay);
            SortedList<DateTime, AmountWithDeviation> outflowDayForecast = Account.GetZeroDayForecast(firstDay, lastDay);
            this.InflowAccount.ComputeDayForecast(firstDay, lastDay, inflowDayForecast);
            this.OutflowAccount.ComputeDayForecast(firstDay, lastDay, outflowDayForecast);
            DateTime date = firstDay;
            while (date <= lastDay)
            {
                if (!dayForecast.ContainsKey(date)) { dayForecast[date] = AmountWithDeviation.Zero; }
                dayForecast[date] += inflowDayForecast[date] - outflowDayForecast[date] + this.InitialBalance;
                date = date.AddTicks(TimeSpan.TicksPerDay);
            }
        }

        /// <summary>
        /// Computes the explanation of the day forecast, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </summary>
        /// <param name="firstDay">First day of the day forecast explanation.</param>
        /// <param name="lastDay">Last day of the day forecast explanation.</param>
        /// <param name="dayForecastExplanation">
        /// Explanation of the day forecast to be computed, mapping every day in the given range 
        /// to the AmountExplanation structure representing the explanation of the total change 
        /// of the account balance at the specified date.
        /// </param>
        /// <remarks>
        /// For technical reasons, we assume that the dayForecastExplanation parameter is already initialized and empty.
        /// Otherwise, if it contains some values, the new computed values will be just added to the stored values.
        /// </remarks>
        public override void ExplainDayForcast(DateTime firstDay, DateTime lastDay, SortedList<DateTime, AmountExplanation> dayForecastExplanation)
        {
            firstDay = firstDay.Date;
            lastDay = lastDay.Date;
            SortedList<DateTime, AmountExplanation> inflowDayForecastExplanation = Account.GetZeroDayExplanation(firstDay, lastDay);
            SortedList<DateTime, AmountExplanation> outflowDayForecastExplanation = Account.GetZeroDayExplanation(firstDay, lastDay);
            this.InflowAccount.ExplainDayForcast(firstDay, lastDay, inflowDayForecastExplanation);
            this.OutflowAccount.ExplainDayForcast(firstDay, lastDay, outflowDayForecastExplanation);
            DateTime date = firstDay;
            while (date <= lastDay)
            {
                if (!dayForecastExplanation.ContainsKey(date)) { dayForecastExplanation[date] = AmountExplanation.Zero; }
                dayForecastExplanation[date] += inflowDayForecastExplanation[date] - outflowDayForecastExplanation[date];
                date = date.AddTicks(TimeSpan.TicksPerDay);
            }
        }

        /// <summary>
        /// Set a new start date for the account.
        /// </summary>
        /// <param name="date">Start date of the account.</param>
        public override void AdjustStartDate(DateTime date)
        {
            AmountWithDeviation oldBalance = this.ExpectedBalance(date);
            this.InflowAccount.AdjustStartDate(date);
            this.OutflowAccount.AdjustStartDate(date);
            AmountWithDeviation newBalance = this.ExpectedBalance(date);
            this.InitialBalance += oldBalance.Amount - newBalance.Amount;
        }

        /// <summary>
        /// Returns the constant account balance of this account in the case when the account balance 
        /// of this account does not change from the given specified date. Otherwise, returns null.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether this account is constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the constant account balance of this account 
        /// in the case when the account balance of this account does not change from the given specified date. 
        /// Otherwise, null.
        /// </returns>
        public override AmountWithDeviation? IsConstant(DateTime fromDate)
        {
            return null;
        }

        /// <summary>
        /// Clears all the constant subaccounts of this account, that are constant from the given specified date, 
        /// and returns the total sum of their balances.
        /// </summary>
        /// <param name="fromDate">Date, from which we measure whether the subaccounts are constant or not.</param>
        /// <returns>
        /// AmountWithDeviation structure representing the total sum of balances of all constant subaccounts 
        /// of this account, which are constant from the given specified date.
        /// </returns>
        public override AmountWithDeviation ClearConstantSubAccounts(DateTime fromDate)
        {
            AmountWithDeviation inflowAmount = this.InflowAccount.ClearConstantSubAccounts(fromDate);
            AmountWithDeviation outflowAmount = this.OutflowAccount.ClearConstantSubAccounts(fromDate);
            this.InitialBalance += (inflowAmount - outflowAmount).Amount;
            return AmountWithDeviation.Zero;
        }

        /// <summary>
        /// Sorts the account.
        /// </summary>
        public override void Sort()
        {
            this.InflowAccount.Sort();
            this.OutflowAccount.Sort();
        }
    }
}
