﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Xml.Serialization;
using CashCalc.Common;

namespace CashCalc.Base
{
    /// <summary>
    /// Account whose balance changes every week by a constant amount.
    /// </summary>
    [Serializable]
    [XmlRoot("WeekAccount")]
    public sealed class WeekAccount : EnumerableAccount
    {
        /// <summary>
        /// Default constructor of the WeekAccount class (due to the XML serialization).
        /// </summary>
        public WeekAccount()
            : this(System.DateTime.Today, System.DateTime.Today.AddYears(1), 0, 0.0, 0)
        {

        }

        /// <summary>
        /// Constructor of the WeekAccount class.
        /// </summary>
        /// <param name="startDate">First date when the account can (possibly) change.</param>
        /// <param name="endDate">Last date when the account can (possibly) change.</param>
        /// <param name="amount">Amount of change of the account balance.</param>
        /// <param name="amountDeviation">Standard deviation of the change.</param>
        /// <param name="selectedDays">Selected days of the week when the account balance changes by a constant amount.</param>
        public WeekAccount(DateTime startDate, DateTime endDate, decimal amount, double amountDeviation, DayFlags selectedDays)
            : this(startDate, endDate, amount, amountDeviation, selectedDays, null)
        {

        }

        /// <summary>
        /// Constructor of the WeekAccount class.
        /// </summary>
        /// <param name="startDate">First date when the account can (possibly) change.</param>
        /// <param name="endDate">Last date when the account can (possibly) change.</param>
        /// <param name="amount">Amount of change of the account balance.</param>
        /// <param name="amountDeviation">Standard deviation of the change.</param>
        /// <param name="selectedDays">Selected days of the week when the account balance changes by a constant amount.</param>
        /// <param name="exceptions">Exceptions.</param>
        public WeekAccount(DateTime startDate, DateTime endDate, decimal amount, double amountDeviation, DayFlags selectedDays, SerializableSortedList<DateTime, AmountWithDeviation> exceptions)
        {
            this.Title = "Week Account";
            this.StartDate = startDate.Date;
            this.EndDate = endDate.Date;
            this.Amount = new AmountWithDeviation
            {
                Amount = amount,
                AmountDeviation = amountDeviation
            };
            this.SelectedDays = selectedDays;
            this.Exceptions = exceptions;
        }

        /// <summary>
        /// Enumeration of days of the week.
        /// </summary>
        [Flags]
        public enum DayFlags
        {
            /// <summary>
            /// Flag for Monday.
            /// </summary>
            Monday = 1,

            /// <summary>
            /// Flag for Tuesday.
            /// </summary>
            Tuesday = 2,

            /// <summary>
            /// Flag for Wednesday.
            /// </summary>
            Wednesday = 4,

            /// <summary>
            /// Flag for Thursday.
            /// </summary>
            Thursday = 8,

            /// <summary>
            /// Flag fo Friday.
            /// </summary>
            Friday = 16,

            /// <summary>
            /// Flag for Saturday.
            /// </summary>
            Saturday = 32,

            /// <summary>
            /// Flag for Sunday.
            /// </summary>
            Sunday = 64
        }

        /// <summary>
        /// Selected days of the week when the account balance changes by a constant amount.
        /// </summary>
        public DayFlags SelectedDays { get; set; }

        /// <summary>
        /// Set a new start date for the account.
        /// </summary>
        /// <param name="date">Start date of the account.</param>
        public override void AdjustStartDate(DateTime date)
        {
            if (date.Date > this.StartDate.Date)
            {
                this.StartDate = date.Date;
            }
        }

        /// <summary>
        /// Enumeration of all dates when the account balance changes by a constant amount.
        /// </summary>
        public override IEnumerable<DateTime> Dates
        {
            get
            {                
                DateTime date = this.StartDate.Date;
                while (date <= this.EndDate)
                {
                    DayFlags dayFlag = this.GetDayFlag(date);
                    if ((this.SelectedDays & dayFlag) == dayFlag)
                    {
                        yield return date;
                    }
                    date = date.AddTicks(TimeSpan.TicksPerDay);
                }
            }
        }

        /// <summary>
        /// Returns a day flag corresponding to the given date.
        /// </summary>
        /// <param name="date">Input date.</param>
        /// <returns>Day flag.</returns>
        private DayFlags GetDayFlag(DateTime date)
        {
            switch (date.DayOfWeek)
            {
                case DayOfWeek.Monday: return DayFlags.Monday;
                case DayOfWeek.Tuesday: return DayFlags.Tuesday;
                case DayOfWeek.Wednesday: return DayFlags.Wednesday;
                case DayOfWeek.Thursday: return DayFlags.Thursday;
                case DayOfWeek.Friday: return DayFlags.Friday;
                case DayOfWeek.Saturday: return DayFlags.Saturday;
                case DayOfWeek.Sunday: return DayFlags.Sunday;
                default: throw new InvalidProgramException("Unknown day of week.");
            }
        }
    }
}
