﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Xml.Serialization;
using System.Windows.Forms.DataVisualization.Charting;

using CashCalc.Base;
using CashCalc.Dialogs;

namespace CashCalc
{
    /// <summary>
    /// Main form of the CashCalc application.
    /// </summary>
    public partial class CashCalcForm : Form
    {
        /// <summary>
        /// Default constructor.
        /// </summary>
        public CashCalcForm()
        {
            InitializeComponent();

            this.NewProject();
        }

        private InflowOutflowAccount inflowOutflowAccount = null;
        private SortedList<DateTime, AmountWithDeviation> forecast;
        private SortedList<DateTime, AmountExplanation> inflowForecastExplanation;
        private SortedList<DateTime, AmountExplanation> outflowForecastExplanation;
        private DateTime cursorDateTime;

        AmountExplanation periodInflowExplanation;
        AmountExplanation periodOutflowExplanation;

        private bool enableUpdate = true;
        private bool dirtyFlag = false;
        private string projectName = null;

        private void BeginUpdate()
        {
            this.enableUpdate = false;
        }

        private void EndUpdate()
        {
            this.enableUpdate = true;
            this.UpdateAll();
        }

        private void UpdateAll()
        {
            this.UpdateCaption();
            this.UpdateChart();
            this.UpdateExplanation();
        }

        private InflowOutflowAccount BaseAccount
        {
            get
            {
                return this.inflowOutflowAccount;
            }
            set
            {
                this.inflowOutflowAccount = value;
                this.initialAmoumtNumericUpDown.Value = value.InitialBalance;
                this.InitializeTreeView(true);
                this.UpdateChart();
                this.UpdateExplanation();
            }
        }

        private DateTime FromDate
        {
            get { return this.fromDateTimePicker.Value.Date; }
            set { this.fromDateTimePicker.Value = value.Date; }
        }

        private DateTime ToDate
        {
            get
            {
                if (this.week1RadioButton.Checked) { return this.FromDate.AddTicks(7 * TimeSpan.TicksPerDay); }
                else if (this.month1RadioButton.Checked) { return this.FromDate.AddMonths(1); }
                else if (this.month3RadioButton.Checked) { return this.FromDate.AddMonths(3); }
                else if (this.month6RadioButton.Checked) { return this.FromDate.AddMonths(6); }
                else if (this.year1RadioButton.Checked) { return this.FromDate.AddYears(1); }
                else if (this.year2RadioButton.Checked) { return this.FromDate.AddYears(2); }
                else if (this.year5RadioButton.Checked) { return this.FromDate.AddYears(5); }
                else if (this.toRadioButton.Checked) { return this.toDateTimePicker.Value.AddTicks(TimeSpan.TicksPerDay); }
                else { throw new InvalidProgramException("No radio button is checked"); }
            }
            set
            {
                if (value <= this.FromDate.AddTicks(7 * TimeSpan.TicksPerDay)) { this.week1RadioButton.Checked = true; }
                else if (value <= this.FromDate.AddMonths(1)) { this.month1RadioButton.Checked = true; }
                else if (value <= this.FromDate.AddMonths(3)) { this.month3RadioButton.Checked = true; }
                else if (value <= this.FromDate.AddMonths(6)) { this.month6RadioButton.Checked = true; }
                else if (value <= this.FromDate.AddYears(1)) { this.year1RadioButton.Checked = true; }
                else if (value <= this.FromDate.AddYears(2)) { this.year2RadioButton.Checked = true; }
                else { this.year5RadioButton.Checked = true; }
            }
        }

        private bool IsDayForecast
        {
            get
            {
                return (this.ToDate < this.FromDate.AddYears(1));
            }
        }

        private bool IsMonthForecast
        {
            get
            {
                return (!this.IsDayForecast);
            }
        }

        private bool DirtyFlag
        {
            get { return this.dirtyFlag; }
            set
            {
                this.dirtyFlag = value;
                this.UpdateCaption();
            }
        }

        private string ProjectName
        {
            get { return this.projectName; }
            set
            {
                this.projectName = value;
                this.UpdateCaption();
            }
        }

        private void NewProject()
        {
            this.BeginUpdate();
            this.BaseAccount = new InflowOutflowAccount();
            this.fromDateTimePicker.Value = DateTime.Today;
            this.initialAmoumtNumericUpDown.Value = 0;
            this.week1RadioButton.Checked = true;
            this.seriesChart.ChartAreas["ChartArea"].CursorX.SetCursorPosition(Double.NaN);
            this.cursorDateTime = DateTime.MinValue;
            this.deviationToolStripMenuItem.Checked = true;
            this.volumeToolStripMenuItem.Checked = true;
            this.explanationToolStripMenuItem.Checked = false;
            this.accountsSplitContainer.Panel2Collapsed = true;
            this.DirtyFlag = false;
            this.ProjectName = null;
            this.EndUpdate();
        }

        private System.Windows.Forms.DialogResult PromptNewProjectName()
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "Xml files (*.xml)|*.xml|All files (*.*)|*.*";
            saveFileDialog.RestoreDirectory = true;
            System.Windows.Forms.DialogResult dialogResult = saveFileDialog.ShowDialog();
            if (dialogResult == System.Windows.Forms.DialogResult.OK)
            {
                this.ProjectName = saveFileDialog.FileName;
                return System.Windows.Forms.DialogResult.OK;
            }
            else
            {
                return dialogResult;
            }
        }

        private bool Save()
        {
            if (String.IsNullOrEmpty(this.ProjectName))
            {
                if (this.PromptNewProjectName() != System.Windows.Forms.DialogResult.OK)
                {
                    return false;
                }
            }
            using (Stream saveStream = File.Open(this.ProjectName, FileMode.Create))
            {
                if (saveStream != null)
                {
                    this.SerializeCashCalc(saveStream, true);
                    saveStream.Close();
                    this.DirtyFlag = false;
                    return true;
                }
                else
                {
                    MessageBox.Show("Could not save to the file: " + this.ProjectName, "Error");
                    return false;
                }
            }
        }

        private bool SaveAs()
        {
            if (this.PromptNewProjectName() == System.Windows.Forms.DialogResult.OK)
            {
                return this.Save();
            }
            else
            {
                return false;
            }
        }

        private System.Windows.Forms.DialogResult PromptSave()
        {
            if (this.DirtyFlag)
            {
                System.Windows.Forms.DialogResult confirmationResult =
                    MessageBox.Show("Do you want to save the project?", "Confirmation", MessageBoxButtons.YesNoCancel);
                if (confirmationResult == System.Windows.Forms.DialogResult.Yes)
                {
                    if (this.Save())
                    {
                        return System.Windows.Forms.DialogResult.Yes;
                    }
                    else
                    {
                        return System.Windows.Forms.DialogResult.Cancel;
                    }
                }
                else
                {
                    return confirmationResult;
                }
            }
            else
            {
                return System.Windows.Forms.DialogResult.Yes;
            }
        }

        private void Open()
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "Xml files (*.xml)|*.xml|All files (*.*)|*.*";
            openFileDialog.RestoreDirectory = true;
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                using (Stream openStream = openFileDialog.OpenFile())
                {
                    if (openStream != null)
                    {
                        this.BeginUpdate();
                        this.DeserializeCashCalc(openStream, true);
                        openStream.Close();
                        this.seriesChart.ChartAreas["ChartArea"].CursorX.SetCursorPosition(Double.NaN);
                        this.cursorDateTime = DateTime.MinValue;
                        this.deviationToolStripMenuItem.Checked = true;
                        this.volumeToolStripMenuItem.Checked = true;
                        this.explanationToolStripMenuItem.Checked = false;
                        this.accountsSplitContainer.Panel2Collapsed = true;
                        this.DirtyFlag = false;
                        this.ProjectName = openFileDialog.FileName;
                        this.EndUpdate();
                    }
                    else
                    {
                        MessageBox.Show("Could not open the file: " + openFileDialog.FileName, "Error");
                    }
                }
            }
        }

        private static bool SerializeAccount(TextWriter stream, Account account, bool show)
        {
            try
            {
                XmlSerializer xmlSerializer = new XmlSerializer(typeof(XmlAccount));
                xmlSerializer.Serialize(stream, new XmlAccount(account));
                return true;
            }
            catch (Exception exception)
            {
                if (show)
                {
                    MessageBox.Show("Could not serialize the account due to the following error: " + exception.Message, "Error");
                }
                return false;
            }
        }

        private static Account DeserializeAccount(TextReader stream, bool show)
        {
            try
            {
                XmlSerializer xmlSerializer = new XmlSerializer(typeof(XmlAccount));
                XmlAccount xmlAccount = xmlSerializer.Deserialize(stream) as XmlAccount;
                return xmlAccount.Account;
            }
            catch (Exception exception)
            {
                if (show)
                {
                    MessageBox.Show("Could not deserialize the account due to the following error: " + exception.Message, "Error");
                }
                return null;
            }
        }

        private void SerializeCashCalc(Stream stream, bool show)
        {
            try
            {
                XmlSerializer xmlSerializer = new XmlSerializer(typeof(XmlCashCalc));
                XmlCashCalc xmlCashCalc = new XmlCashCalc(
                    this.inflowOutflowAccount,
                    this.FromDate,
                    this.ToDate);
                xmlSerializer.Serialize(stream, xmlCashCalc);
            }
            catch (Exception exception)
            {
                if (show)
                {
                    MessageBox.Show("Could not serialize the CashCalc project due to the following error: " + exception.Message, "Error");
                }
            }
        }

        private void DeserializeCashCalc(Stream stream, bool show)
        {
            try
            {
                XmlSerializer xmlSerializer = new XmlSerializer(typeof(XmlCashCalc));
                XmlCashCalc xmlCashCalc = xmlSerializer.Deserialize(stream) as XmlCashCalc;
                this.BaseAccount = xmlCashCalc.InflowOutflowAccount;
                this.FromDate = xmlCashCalc.FromDate.Date;
                this.ToDate = xmlCashCalc.ToDate.Date;
            }
            catch (Exception exception)
            {
                if (show)
                {
                    MessageBox.Show("Could not deserialize the CashCalc project due to the following error: " + exception.Message, "Error");
                }
            }
        }

        private void AdjustStartDate(DateTime adjustDate)
        {
            adjustDate = adjustDate.Date;
            System.Windows.Forms.DialogResult confirmation = MessageBox.Show(
                "Do you also want to clear all the constant accounts before the date: " +
                adjustDate.ToString("D", System.Globalization.CultureInfo.InvariantCulture) + "?", 
                "Confirmation", MessageBoxButtons.YesNoCancel);
            if (confirmation == System.Windows.Forms.DialogResult.Yes)
            {
                this.inflowOutflowAccount.ClearConstantSubAccounts(adjustDate);
            }
            if (confirmation != System.Windows.Forms.DialogResult.Cancel)
            {
                this.inflowOutflowAccount.AdjustStartDate(adjustDate);
                this.initialAmoumtNumericUpDown.Value = this.inflowOutflowAccount.InitialBalance;
                this.DirtyFlag = true;
                this.InitializeTreeView(true);
                this.UpdateChart();
                this.UpdateExplanation();
            }
        }

        private void InitializeTreeView(bool recursive)
        {
            this.inflowOutflowTreeView.Nodes.Clear();
            TreeNode inflowNode = this.inflowOutflowTreeView.Nodes.Add("InflowNode");
            TreeNode outflowNode = this.inflowOutflowTreeView.Nodes.Add("OutflowNode");
            UpdateTreeNode(outflowNode, this.inflowOutflowAccount.OutflowAccount, recursive);
            UpdateTreeNode(inflowNode, this.inflowOutflowAccount.InflowAccount, recursive);
            this.Sort();
        }

        private void UpdateCaption()
        {
            if (!this.enableUpdate) { return; }

            if (String.IsNullOrEmpty(this.ProjectName))
            {
                if (this.DirtyFlag) this.Text = "*Untitled - Cash Calculator";
                else this.Text = "Untitled - Cash Calculator";
            }
            else
            {
                if (this.DirtyFlag) this.Text = "*" + this.ProjectName + " - Cash Calculator";
                else this.Text = this.ProjectName + " - Cash Calculator";
            }
        }

        private void UpdateVolumeView()
        {
            this.seriesChart.ChartAreas["ChartAreaBar"].Visible = this.volumeToolStripMenuItem.Checked;
            this.seriesChart.ChartAreas["ChartArea"].AxisX.LabelStyle.Enabled = !this.volumeToolStripMenuItem.Checked;
            if (this.volumeToolStripMenuItem.Checked)
            {
                this.seriesChart.ChartAreas["ChartArea"].Position.Height = 60;
            }
            else
            {
                this.seriesChart.ChartAreas["ChartArea"].Position.Height = 90;
            }
        }

        private void UpdateSeriesChartAxes()
        {
            if (this.ToDate <= this.FromDate.AddTicks(7 * TimeSpan.TicksPerDay))
            {
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartArea"].CursorX.IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.Series["LineSeries"].MarkerStyle = MarkerStyle.Circle;
                this.seriesChart.Series["LineSeries"].MarkerSize = 8;
            }
            else if (this.ToDate <= this.FromDate.AddMonths(1))
            {
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Weeks;
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Weeks;
                this.seriesChart.ChartAreas["ChartArea"].CursorX.IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.Series["LineSeries"].MarkerStyle = MarkerStyle.Circle;
                this.seriesChart.Series["LineSeries"].MarkerSize = 8;
            }
            else if (this.ToDate <= this.FromDate.AddMonths(3))
            {
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Weeks;
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Weeks;
                this.seriesChart.ChartAreas["ChartArea"].CursorX.IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.Series["LineSeries"].MarkerStyle = MarkerStyle.Circle;
                this.seriesChart.Series["LineSeries"].MarkerSize = 4;
            }
            else if (this.ToDate <= this.FromDate.AddMonths(6))
            {
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Weeks;
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Weeks;
                this.seriesChart.ChartAreas["ChartArea"].CursorX.IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Days;
                this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.IntervalType = DateTimeIntervalType.Days;
                this.seriesChart.Series["LineSeries"].MarkerStyle = MarkerStyle.Circle;
                this.seriesChart.Series["LineSeries"].MarkerSize = 4;
            }
            else if (this.ToDate <= this.FromDate.AddYears(1))
            {
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartArea"].CursorX.IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.Series["LineSeries"].MarkerStyle = MarkerStyle.Circle;
                this.seriesChart.Series["LineSeries"].MarkerSize = 8;
            }
            else if (this.ToDate <= this.FromDate.AddYears(2))
            {
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartArea"].CursorX.IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.Series["LineSeries"].MarkerStyle = MarkerStyle.Circle;
                this.seriesChart.Series["LineSeries"].MarkerSize = 8;
            }
            else if (this.ToDate <= this.FromDate.AddYears(5))
            {
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartArea"].CursorX.IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.Series["LineSeries"].MarkerStyle = MarkerStyle.Circle;
                this.seriesChart.Series["LineSeries"].MarkerSize = 4;
            }
            else
            {
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartArea"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartArea"].CursorX.IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].AxisX.StripLines[0].StripWidthType = DateTimeIntervalType.Months;
                this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.IntervalType = DateTimeIntervalType.Months;
                this.seriesChart.Series["LineSeries"].MarkerStyle = MarkerStyle.None;
                this.seriesChart.Series["LineSeries"].MarkerSize = 0;
            }
        }

        private void UpdateSeriesChartCursor()
        {
            this.seriesChart.ChartAreas["ChartArea"].CursorX.IsUserEnabled = this.explanationToolStripMenuItem.Checked;
            this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.IsUserEnabled = this.explanationToolStripMenuItem.Checked;
            if (!this.explanationToolStripMenuItem.Checked)
            {
                this.seriesChart.ChartAreas["ChartArea"].CursorX.SetCursorPosition(0.0);
                this.seriesChart.ChartAreas["ChartAreaBar"].CursorX.SetCursorPosition(0.0);
                this.cursorDateTime = DateTime.MinValue;
            }
        }

        private void BuildExplanationToolTip(StringBuilder toolTipStringBuilder, KeyValuePair<DateTime, AmountExplanation> explanationPair, string title, bool isDayForecast)
        {
            if (isDayForecast)
            {
                toolTipStringBuilder.AppendLine(explanationPair.Key.ToString("D", System.Globalization.CultureInfo.InvariantCulture));
            }
            else
            {
                toolTipStringBuilder.AppendLine(explanationPair.Key.ToString("y", System.Globalization.CultureInfo.InvariantCulture));
            }
            toolTipStringBuilder.Append(title);
            toolTipStringBuilder.AppendLine(explanationPair.Value.Amount.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
            foreach (Tuple<Account, AmountWithDeviation> explanation in explanationPair.Value.Explanation.OrderBy(touple => touple.Item2.Amount).Reverse())
            {
                toolTipStringBuilder.Append("  ");
                toolTipStringBuilder.Append(explanation.Item1.Title);
                toolTipStringBuilder.Append(": ");
                toolTipStringBuilder.AppendLine(explanation.Item2.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                //toolTipStringBuilder.Append(" +- ");
                //toolTipStringBuilder.AppendLine(explanation.Item2.AmountDeviation.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
            }
        }

        private void UpdateSeriesChart()
        {
            DateTime firstDay = this.FromDate;
            DateTime lastDay = this.ToDate.AddTicks(-TimeSpan.TicksPerDay);
            DateTime firstMonth = new DateTime(firstDay.Year, firstDay.Month, 1);
            DateTime lastMonth = new DateTime(lastDay.Year, lastDay.Month, 1);

            if (this.IsDayForecast)
            {
                this.forecast = Account.GetZeroDayForecast(firstDay, lastDay);
                this.inflowForecastExplanation = Account.GetZeroDayExplanation(firstDay, lastDay);
                this.outflowForecastExplanation = Account.GetZeroDayExplanation(firstDay, lastDay);

                this.inflowOutflowAccount.ComputeDayForecast(firstDay, lastDay, forecast);
                this.inflowOutflowAccount.InflowAccount.ExplainDayForcast(firstDay, lastDay, inflowForecastExplanation);
                this.inflowOutflowAccount.OutflowAccount.ExplainDayForcast(firstDay, lastDay, outflowForecastExplanation);
            }
            else
            {
                this.forecast = Account.GetZeroMonthForecast(firstMonth, lastMonth);
                this.inflowForecastExplanation = Account.GetZeroMonthExplanation(firstMonth, lastMonth);
                this.outflowForecastExplanation = Account.GetZeroMonthExplanation(firstMonth, lastMonth);

                this.inflowOutflowAccount.ComputeMonthForecast(firstMonth, lastMonth, forecast);
                this.inflowOutflowAccount.InflowAccount.ExplainMonthForcast(firstMonth, lastMonth, inflowForecastExplanation);
                this.inflowOutflowAccount.OutflowAccount.ExplainMonthForcast(firstMonth, lastMonth, outflowForecastExplanation);
            }

            this.seriesChart.SuspendLayout();

            this.UpdateSeriesChartAxes();

            this.UpdateSeriesChartCursor();

            this.UpdateVolumeView();

            Series lineSeries = this.seriesChart.Series["LineSeries"];
            Series rangeSeries = this.seriesChart.Series["RangeSeries"];
            Series inflowSeries = this.seriesChart.Series["BarSeriesInflow"];
            Series outflowSeries = this.seriesChart.Series["BarSeriesOutflow"];

            lineSeries.Points.Clear();
            rangeSeries.Points.Clear();
            inflowSeries.Points.Clear();
            outflowSeries.Points.Clear();

            StringBuilder toolTipStringBuilder = new StringBuilder();

            bool isDayForecast = this.IsDayForecast;

            foreach (KeyValuePair<DateTime, AmountWithDeviation> forecastPair in this.forecast)
            {
                DataPoint lineDataPoint = new DataPoint(forecastPair.Key.ToOADate(), (double)forecastPair.Value.Amount);
                toolTipStringBuilder.Clear();
                if (isDayForecast)
                {
                    toolTipStringBuilder.AppendLine(forecastPair.Key.ToString("D", System.Globalization.CultureInfo.InvariantCulture));
                }
                else
                {
                    DateTime lastMonthDate = new DateTime(forecastPair.Key.Year, forecastPair.Key.Month, 1).AddMonths(1).AddTicks(-TimeSpan.TicksPerDay).Date;
                    toolTipStringBuilder.AppendLine(lastMonthDate.ToString("D", System.Globalization.CultureInfo.InvariantCulture));
                }
                toolTipStringBuilder.Append("Balance: ");
                toolTipStringBuilder.Append(forecastPair.Value.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                //toolTipStringBuilder.Append(" +- ");
                //toolTipStringBuilder.Append(forecastPair.Value.AmountDeviation.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                lineDataPoint.ToolTip = toolTipStringBuilder.ToString();
                lineSeries.Points.Add(lineDataPoint);
                rangeSeries.Points.AddXY(forecastPair.Key,
                    (double)forecastPair.Value.Amount - forecastPair.Value.AmountDeviation,
                    (double)forecastPair.Value.Amount + forecastPair.Value.AmountDeviation);
            }

            foreach (KeyValuePair<DateTime, AmountExplanation> forecastExplanationPair in this.inflowForecastExplanation)
            {
                DataPoint inflowDataPoint = new DataPoint(forecastExplanationPair.Key.ToOADate(), (double)forecastExplanationPair.Value.Amount.Amount);
                toolTipStringBuilder.Clear();
                this.BuildExplanationToolTip(toolTipStringBuilder, forecastExplanationPair, "Inflow: ", isDayForecast);
                inflowDataPoint.ToolTip = toolTipStringBuilder.ToString();
                inflowSeries.Points.Add(inflowDataPoint);
            }

            foreach (KeyValuePair<DateTime, AmountExplanation> forecastExplanationPair in this.outflowForecastExplanation)
            {
                DataPoint outflowDataPoint = new DataPoint(forecastExplanationPair.Key.ToOADate(), (double)forecastExplanationPair.Value.Amount.Amount);
                toolTipStringBuilder.Clear();
                this.BuildExplanationToolTip(toolTipStringBuilder, forecastExplanationPair, "Outflow: ", isDayForecast);
                outflowDataPoint.ToolTip = toolTipStringBuilder.ToString();
                outflowSeries.Points.Add(outflowDataPoint);
            }

            this.seriesChart.ResumeLayout();
        }

        private void UpdatePieSeries(Series series, AmountExplanation explanation)
        {
            series.Points.Clear();
            if (explanation.Amount.Amount != 0 && explanation.Explanation.Count > 0)
            {
                foreach (Tuple<Account, AmountWithDeviation> explanationPair in explanation.Explanation.OrderBy(touple => touple.Item2.Amount).Reverse())
                {
                    double percent = 100.0 * (double)(explanationPair.Item2.Amount / explanation.Amount.Amount);
                    DataPoint dataPoint = new DataPoint(0, percent);
                    if (percent > 10.0)
                    {
                        dataPoint.Label = percent.ToString("F", System.Globalization.CultureInfo.InvariantCulture) + "%";
                    }
                    else
                    {
                        dataPoint.Label = String.Empty;
                    }
                    dataPoint.LegendText = explanationPair.Item1.Title + ": " + explanationPair.Item2.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture) +
                        " [" + percent.ToString("F", System.Globalization.CultureInfo.InvariantCulture) + "%]";
                    dataPoint.ToolTip = dataPoint.LegendText;
                    series.Points.Add(dataPoint);
                }
            }
        }

        private void UpdatePieChart()
        {
            DateTime firstDay = this.FromDate;
            DateTime lastDay = this.ToDate.AddTicks(-TimeSpan.TicksPerDay);

            this.pieChart.SuspendLayout();

            this.pieChart.ChartAreas["InflowChartArea"].Visible = this.inflowToolStripMenuItem.Checked;
            this.pieChart.ChartAreas["OutflowChartArea"].Visible = this.outflowToolStripMenuItem.Checked;

            this.periodInflowExplanation = this.inflowOutflowAccount.InflowAccount.ExplainPeriod(firstDay, lastDay);
            this.periodOutflowExplanation = this.inflowOutflowAccount.OutflowAccount.ExplainPeriod(firstDay, lastDay);

            this.UpdatePieSeries(this.pieChart.Series["InflowPie"], this.periodInflowExplanation);
            this.UpdatePieSeries(this.pieChart.Series["OutflowPie"], this.periodOutflowExplanation);

            string periodDateString = firstDay.ToString("d", System.Globalization.CultureInfo.InvariantCulture) + " - " +
                lastDay.ToString("d", System.Globalization.CultureInfo.InvariantCulture);

            this.pieChart.Titles[0].Text = "Inflow [" + periodDateString + "]: " + this.periodInflowExplanation.Amount.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture);
            this.pieChart.Titles[1].Text = "Outflow [" + periodDateString + "]: " + this.periodOutflowExplanation.Amount.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture);

            this.pieChart.ResumeLayout();
        }
        
        private void UpdateChart()
        {
            if (!this.enableUpdate) { return; }

            if (this.tabControl.SelectedIndex == 0)
            {
                this.UpdateSeriesChart();
            }
            else if (this.tabControl.SelectedIndex == 1)
            {
                this.UpdatePieChart();
            }
        }

        private void UpdateExplanation()
        {
            if (!this.enableUpdate) { return; }

            string balanceString = String.Empty;
            string accountPrefix = String.Empty;
            DateTime date = DateTime.MinValue;
            if (this.cursorDateTime > DateTime.MinValue)
            {
                if (this.IsDayForecast)
                {
                    date = this.cursorDateTime.Date;
                    this.dateTextBox.Text = date.ToString("D", System.Globalization.CultureInfo.InvariantCulture);
                    balanceString = "Balance [" + date.ToString("d MMM yyyy", System.Globalization.CultureInfo.InvariantCulture) + "]";
                    accountPrefix = "Day ";
                }
                else
                {
                    date = new DateTime(this.cursorDateTime.Year, this.cursorDateTime.Month, 1);
                    this.dateTextBox.Text = date.ToString("y", System.Globalization.CultureInfo.InvariantCulture);
                    DateTime lastMonthDay = date.AddMonths(1).AddTicks(-TimeSpan.TicksPerDay).Date;
                    balanceString = "Balance [" + lastMonthDay.ToString("d MMM yyyy", System.Globalization.CultureInfo.InvariantCulture) + "]";
                    accountPrefix = "Month ";
                }
            }
            else
            {
                this.dateTextBox.Text = String.Empty;
            }
            this.explanationListView.BeginUpdate();
            this.explanationListView.Items.Clear();
            if (this.forecast.ContainsKey(date))
            {
                AmountWithDeviation amount = this.forecast[date];
                ListViewItem item = this.explanationListView.Items.Add(balanceString);
                item.SubItems.Add(amount.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                item.SubItems.Add(amount.AmountDeviation.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
            }
            if (this.inflowForecastExplanation.ContainsKey(date))
            {
                AmountExplanation inflowExplanation = this.inflowForecastExplanation[date];
                ListViewItem item = this.explanationListView.Items.Add(accountPrefix + "Inflow:");
                item.SubItems.Add(inflowExplanation.Amount.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                item.SubItems.Add(inflowExplanation.Amount.AmountDeviation.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                foreach (Tuple<Account, AmountWithDeviation> explanation in inflowExplanation.Explanation.OrderBy(touple => touple.Item2.Amount).Reverse())
                {
                    item = this.explanationListView.Items.Add("  " + explanation.Item1.Title);
                    item.SubItems.Add(explanation.Item2.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                    item.SubItems.Add(explanation.Item2.AmountDeviation.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                }
            }
            if (this.outflowForecastExplanation.ContainsKey(date))
            {
                AmountExplanation outflowExplanation = this.outflowForecastExplanation[date];
                ListViewItem item = this.explanationListView.Items.Add(accountPrefix + "Outflow:");
                item.SubItems.Add(outflowExplanation.Amount.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                item.SubItems.Add(outflowExplanation.Amount.AmountDeviation.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                foreach (Tuple<Account, AmountWithDeviation> explanation in outflowExplanation.Explanation.OrderBy(touple => touple.Item2.Amount).Reverse())
                {
                    item = this.explanationListView.Items.Add("  " + explanation.Item1.Title);
                    item.SubItems.Add(explanation.Item2.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                    item.SubItems.Add(explanation.Item2.AmountDeviation.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                }
            }
            this.explanationListView.EndUpdate();
        }

        private void Sort()
        {
            this.inflowOutflowAccount.Sort();
            this.inflowOutflowTreeView.Sort();
        }

        private static CheckBoxAccount GetCheckBoxAccount(TreeNode node)
        {
            if (node != null && node.Tag is CheckBoxAccount)
                return node.Tag as CheckBoxAccount;
            else return null;
        }

        private static Account GetInnerAccount(TreeNode node)
        {
            CheckBoxAccount checkBoxAccount = GetCheckBoxAccount(node);
            if (checkBoxAccount != null)
                return checkBoxAccount.Account;
            else return null;
        }

        private void UpdateTreeNode(TreeNode node, CheckBoxAccount checkBoxAccount)
        {
            UpdateTreeNode(node, checkBoxAccount, false);
        }

        private void UpdateTreeNode(TreeNode node, CheckBoxAccount checkBoxAccount, bool recursive)
        {
            node.Tag = null;
            node.Text = checkBoxAccount.ToString();
            node.Checked = checkBoxAccount.Enabled;
            if (recursive && checkBoxAccount.Account is CompositeAccount)
            {
                CompositeAccount compositeAccount = checkBoxAccount.Account as CompositeAccount;
                foreach (CheckBoxAccount childCheckBoxAccount in compositeAccount.CheckBoxAccounts)
                {
                    TreeNode childNode = node.Nodes.Add(childCheckBoxAccount.ToString());
                    UpdateTreeNode(childNode, childCheckBoxAccount, recursive);
                }
            }
            node.Tag = checkBoxAccount;
        }

        private void inflowOutflowTreeView_MouseDown(object sender, MouseEventArgs e)
        {
            TreeNode node = this.inflowOutflowTreeView.GetNodeAt(e.Location);
            if (node != null)
            {
                this.inflowOutflowTreeView.SelectedNode = node;
            }
        }

        private void inflowOutflowTreeView_AfterSelect(object sender, TreeViewEventArgs e)
        {
            Account account = GetInnerAccount(e.Node);
            if (account is CompositeAccount)
                this.inflowOutflowTreeView.ContextMenuStrip = this.compositeAccountContextMenuStrip;
            else if (account is Account)
                this.inflowOutflowTreeView.ContextMenuStrip = this.accountContextMenuStrip;
            else this.inflowOutflowTreeView.ContextMenuStrip = null;
        }

        private void inflowOutflowTreeView_AfterCheck(object sender, TreeViewEventArgs e)
        {
            TreeNode treeNode = e.Node;
            if (treeNode != null && treeNode.Tag is CheckBoxAccount)
            {
                CheckBoxAccount checkBoxAccount = treeNode.Tag as CheckBoxAccount;
                checkBoxAccount.Enabled = treeNode.Checked;
                this.UpdateChart();
                this.UpdateExplanation();
            }
        }

        private void addAccountToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TreeNode selectedTreeNode = this.inflowOutflowTreeView.SelectedNode;
            CompositeAccount selectedCompositeAccount = GetInnerAccount(selectedTreeNode) as CompositeAccount;
            AccountForm accountForm = new AccountForm();
            if (accountForm.ShowDialog() == DialogResult.OK)
            {
                CheckBoxAccount newCheckBoxAccount = new CheckBoxAccount(accountForm.Account);
                selectedCompositeAccount.CheckBoxAccounts.Add(newCheckBoxAccount);
                TreeNode newTreeNode = selectedTreeNode.Nodes.Add(newCheckBoxAccount.Title);
                UpdateTreeNode(newTreeNode, newCheckBoxAccount);
                this.Sort();
                this.inflowOutflowTreeView.SelectedNode = newTreeNode;
                this.DirtyFlag = true;
                this.UpdateChart();
                this.UpdateExplanation();
            }
        }

        private void addFolderToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TreeNode selectedTreeNode = this.inflowOutflowTreeView.SelectedNode;
            CompositeAccount selectedCompositeAccount = GetInnerAccount(selectedTreeNode) as CompositeAccount;
            TitleForm titleForm = new TitleForm();
            if (titleForm.ShowDialog() == DialogResult.OK)
            {
                CheckBoxAccount newCheckBoxAccount = new CheckBoxAccount(new CompositeAccount());
                newCheckBoxAccount.Title = titleForm.Title;
                selectedCompositeAccount.CheckBoxAccounts.Add(newCheckBoxAccount);
                TreeNode newTreeNode = selectedTreeNode.Nodes.Add("[" + newCheckBoxAccount.Title + "]");
                UpdateTreeNode(newTreeNode, newCheckBoxAccount);
                this.Sort();
                this.inflowOutflowTreeView.SelectedNode = newTreeNode;
                this.DirtyFlag = true;
            }
        }

        private void propertiesAccountToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TreeNode selectedTreeNode = this.inflowOutflowTreeView.SelectedNode;
            TreeNode parentTreeNode = selectedTreeNode.Parent;
            if (parentTreeNode != null)
            {
                CheckBoxAccount selectedCheckBoxAccount = GetCheckBoxAccount(selectedTreeNode);
                CompositeAccount parentCompositeAccount = GetInnerAccount(parentTreeNode) as CompositeAccount;
                int index = parentCompositeAccount.CheckBoxAccounts.FindIndex(a => a == selectedCheckBoxAccount);
                AccountForm accountForm = new AccountForm();
                accountForm.Account = selectedCheckBoxAccount.Account;
                if (accountForm.ShowDialog() == DialogResult.OK)
                {
                    CheckBoxAccount newCheckBoxAccount = new CheckBoxAccount(accountForm.Account);
                    parentCompositeAccount.CheckBoxAccounts[index] = newCheckBoxAccount;
                    UpdateTreeNode(selectedTreeNode, newCheckBoxAccount);
                    this.Sort();
                    this.DirtyFlag = true;
                    this.UpdateChart();
                    this.UpdateExplanation();
                }
            }
        }

        private void deleteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TreeNode selectedTreeNode = this.inflowOutflowTreeView.SelectedNode;
            TreeNode parentTreeNode = selectedTreeNode.Parent;
            if (parentTreeNode != null)
            {
                if (MessageBox.Show("Are you sure you want to delete this account?", 
                    "Confirm delete", MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    CheckBoxAccount selectedCheckBoxAccount = GetCheckBoxAccount(selectedTreeNode);
                    CompositeAccount parentCompositeAccount = GetInnerAccount(parentTreeNode) as CompositeAccount;
                    parentCompositeAccount.CheckBoxAccounts.Remove(selectedCheckBoxAccount);
                    parentTreeNode.Nodes.Remove(selectedTreeNode);
                    this.DirtyFlag = true;
                    this.UpdateChart();
                    this.UpdateExplanation();
                }
            }
            else MessageBox.Show("Cannot delete the root account.", "Warning");
        }

        private void renameToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TreeNode selectedTreeNode = this.inflowOutflowTreeView.SelectedNode;
            CheckBoxAccount selectedCheckBoxAccount = GetCheckBoxAccount(selectedTreeNode);
            TitleForm titleForm = new TitleForm();
            titleForm.Title = selectedCheckBoxAccount.Title;
            if (titleForm.ShowDialog() == DialogResult.OK)
            {
                selectedCheckBoxAccount.Title = titleForm.Title;
                UpdateTreeNode(selectedTreeNode, selectedCheckBoxAccount);
                this.Sort();
                this.DirtyFlag = true;
            }
        }

        private void cutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TreeNode selectedTreeNode = this.inflowOutflowTreeView.SelectedNode;
            TreeNode parentTreeNode = selectedTreeNode.Parent;
            if (parentTreeNode != null)
            {
                CheckBoxAccount selectedCheckBoxAccount = GetCheckBoxAccount(selectedTreeNode);
                CompositeAccount parentCompositeAccount = GetInnerAccount(parentTreeNode) as CompositeAccount;
                using (StringWriter stringWriter = new StringWriter())
                {
                    if (SerializeAccount(stringWriter, selectedCheckBoxAccount, true))
                    {
                        Clipboard.SetText(stringWriter.ToString());
                    }
                }
                parentCompositeAccount.CheckBoxAccounts.Remove(selectedCheckBoxAccount);
                parentTreeNode.Nodes.Remove(selectedTreeNode);
                this.DirtyFlag = true;
                this.UpdateChart();
                this.UpdateExplanation();
            }
        }

        private void copyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TreeNode selectedTreeNode = this.inflowOutflowTreeView.SelectedNode;
            TreeNode parentTreeNode = selectedTreeNode.Parent;
            if (parentTreeNode != null)
            {
                CheckBoxAccount selectedCheckBoxAccount = GetCheckBoxAccount(selectedTreeNode);
                CompositeAccount parentCompositeAccount = GetInnerAccount(parentTreeNode) as CompositeAccount;
                using (StringWriter stringWriter = new StringWriter())
                {
                    if (SerializeAccount(stringWriter, selectedCheckBoxAccount, true))
                    {
                        Clipboard.SetText(stringWriter.ToString());
                    }
                }
            }
        }

        private void pasteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            TreeNode selectedTreeNode = this.inflowOutflowTreeView.SelectedNode;
            if (selectedTreeNode != null && Clipboard.ContainsText())
            {
                CompositeAccount selectedCompositeAccount = GetInnerAccount(selectedTreeNode) as CompositeAccount;
                using (StringReader stringReader = new StringReader(Clipboard.GetText()))
                {
                    Account account = DeserializeAccount(stringReader, true);
                    if (account is CheckBoxAccount)
                    {
                        CheckBoxAccount checkBoxAccount = account as CheckBoxAccount;
                        selectedCompositeAccount.CheckBoxAccounts.Add(checkBoxAccount);
                        TreeNode newTreeNode = selectedTreeNode.Nodes.Add(checkBoxAccount.Title);
                        UpdateTreeNode(newTreeNode, checkBoxAccount);
                        this.Sort();
                        this.inflowOutflowTreeView.SelectedNode = newTreeNode;
                        this.DirtyFlag = true;
                        this.UpdateChart();
                        this.UpdateExplanation();
                    }
                }
            }
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (this.PromptSave() != System.Windows.Forms.DialogResult.Cancel)
            {
                this.NewProject();
            }
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (this.PromptSave() != System.Windows.Forms.DialogResult.Cancel)
            {
                this.Open();
            }
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Save();
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.SaveAs();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutForm aboutForm = new AboutForm();
            aboutForm.ShowDialog();
        }

        private void fromDateTimePicker_ValueChanged(object sender, EventArgs e)
        {
            this.toDateTimePicker.MinDate = this.fromDateTimePicker.Value.AddTicks(6 * TimeSpan.TicksPerDay);
            this.UpdateChart();
        }

        private void toDateTimePicker_ValueChanged(object sender, EventArgs e)
        {
            this.UpdateChart();
        }

        private void initialAmoumtNumericUpDown_ValueChanged(object sender, EventArgs e)
        {
            this.inflowOutflowAccount.InitialBalance = this.initialAmoumtNumericUpDown.Value;
            this.DirtyFlag = true;
            this.UpdateChart();
            this.UpdateExplanation();
        }

        private void periodRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            this.toDateTimePicker.Enabled = (sender == this.toRadioButton);
            this.UpdateChart();
            this.UpdateExplanation();
        }

        private void seriesChart_CursorPositionChanging(object sender, CursorEventArgs e)
        {
            this.cursorDateTime = DateTime.FromOADate(e.NewPosition);
            this.UpdateExplanation();
        }

        private void adjustStartDateButton_Click(object sender, EventArgs e)
        {
            this.AdjustStartDate(this.fromDateTimePicker.Value.Date);
        }

        private void CashCalcForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.UserClosing)
            {
                if (this.DirtyFlag)
                {
                    if (this.PromptSave() != System.Windows.Forms.DialogResult.Cancel)
                    {
                        e.Cancel = false;
                    }
                    else
                    {
                        e.Cancel = true;
                    }
                }
                else
                {
                    e.Cancel = MessageBox.Show("Do you really want to close the Cash Calculator application?",
                        "Confirmation", MessageBoxButtons.YesNo) == DialogResult.No;
                }
            }
        }

        private void deviationToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.deviationToolStripMenuItem.Checked = !this.deviationToolStripMenuItem.Checked;
            this.viewdeviationToolStripMenuItem.Checked = this.deviationToolStripMenuItem.Checked;
            this.seriesChart.Series["RangeSeries"].Enabled = this.deviationToolStripMenuItem.Checked;
        }

        private void volumeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.volumeToolStripMenuItem.Checked = !this.volumeToolStripMenuItem.Checked;
            this.viewvolumeToolStripMenuItem.Checked = this.volumeToolStripMenuItem.Checked;
            this.UpdateVolumeView();
        }

        private void explanationToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.explanationToolStripMenuItem.Checked = !this.explanationToolStripMenuItem.Checked;
            this.viewexplanationToolStripMenuItem.Checked = this.explanationToolStripMenuItem.Checked;
            this.accountsSplitContainer.Panel2Collapsed = !this.explanationToolStripMenuItem.Checked;
            if (!this.explanationToolStripMenuItem.Checked)
            {
                this.cursorDateTime = DateTime.MinValue;
            }
            this.UpdateChart();
            this.UpdateExplanation();
        }

        private void inflowToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (this.outflowToolStripMenuItem.Checked)
            {
                this.inflowToolStripMenuItem.Checked = !this.inflowToolStripMenuItem.Checked;
                this.viewInflowToolStripMenuItem.Checked = this.inflowToolStripMenuItem.Checked;
                this.UpdateChart();
            }
        }

        private void outflowToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (this.inflowToolStripMenuItem.Checked)
            {
                this.outflowToolStripMenuItem.Checked = !this.outflowToolStripMenuItem.Checked;
                this.viewOutflowToolStripMenuItem.Checked = this.outflowToolStripMenuItem.Checked;
                this.UpdateChart();
            }
        }

        private void exportTocsvToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (this.tabControl.SelectedIndex == 0)
            {
                if (this.forecast != null && this.forecast.Count > 0)
                {
                    ExportSeriesToCsvForm exportSeriesToCsvForm = new ExportSeriesToCsvForm(
                        this.forecast,
                        this.inflowForecastExplanation,
                        this.outflowForecastExplanation,
                        this.IsMonthForecast);
                    exportSeriesToCsvForm.ShowDialog();
                }
            }
            else if (this.tabControl.SelectedIndex == 1)
            {
                ExportInflowOutflowToCsvForm exportInflowOutflowToCsvForm = new ExportInflowOutflowToCsvForm(
                    this.periodInflowExplanation,
                    this.periodOutflowExplanation);
                exportInflowOutflowToCsvForm.ShowDialog();
            }
        }

        private void exportToimageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (this.tabControl.SelectedIndex == 0)
            {
                if (this.forecast != null && this.forecast.Count > 0)
                {
                    ExportToImageForm exportToImageForm = new ExportToImageForm(this.seriesChart);
                    exportToImageForm.ShowDialog();
                }
            }
            else if (this.tabControl.SelectedIndex == 1)
            {
                ExportToImageForm exportToImageForm = new ExportToImageForm(this.pieChart);
                exportToImageForm.ShowDialog();
            }
        }

        private void tabControl_SelectedIndexChanged(object sender, EventArgs e)
        {
            this.UpdateChart();
        }
    }
}
