﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using CashCalc.Base;

namespace CashCalc.Dialogs
{
    /// <summary>
    /// Form for exporting the inflow/outflow diagram to the csv file.
    /// </summary>
    public partial class ExportInflowOutflowToCsvForm : Form
    {
        /// <summary>
        /// Default constructor.
        /// </summary>
        public ExportInflowOutflowToCsvForm()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Constructor with specified inflow and outflow explanations.
        /// </summary>
        /// <param name="periodInflowExplanation">Inflow explanation.</param>
        /// <param name="periodOutflowExplanation">Outflow explanation.</param>
        public ExportInflowOutflowToCsvForm(AmountExplanation periodInflowExplanation, AmountExplanation periodOutflowExplanation)
            : this()
        {
            this.periodInflowExplanation = periodInflowExplanation;
            this.periodOutflowExplanation = periodOutflowExplanation;
        }

        private AmountExplanation periodInflowExplanation;
        private AmountExplanation periodOutflowExplanation;

        private string separator = ";";

        private bool IsInflowChecked
        {
            get
            {
                return this.inflowRadioButton.Checked || this.bothRadioButton.Checked;
            }
        }

        private bool IsOutflowChecked
        {
            get
            {
                return this.outflowRadioButton.Checked || this.bothRadioButton.Checked;
            }
        }

        private void BuildCSV(StringBuilder stringBuilder, AmountExplanation explanation, string title, System.Globalization.CultureInfo culture)
        {
            stringBuilder.Append(title);
            stringBuilder.Append(this.separator);
            stringBuilder.Append(explanation.Amount.Amount.ToString("G", culture));
            stringBuilder.Append(this.separator);
            if (this.deviationCheckBox.Checked)
            {
                stringBuilder.Append(explanation.Amount.AmountDeviation.ToString("G", culture));
                stringBuilder.Append(this.separator);
            }
            if (this.percentCheckBox.Checked)
            {
                stringBuilder.Append(((double)100.0).ToString("G", culture));
                stringBuilder.Append(this.separator);
            }
            stringBuilder.AppendLine();
            foreach (Tuple<Account, AmountWithDeviation> explanationPair in explanation.Explanation.OrderBy(touple => touple.Item2.Amount).Reverse())
            {
                stringBuilder.Append(explanationPair.Item1.Title);
                stringBuilder.Append(this.separator);
                stringBuilder.Append(explanationPair.Item2.Amount.ToString("G", culture));
                stringBuilder.Append(this.separator);
                if (this.deviationCheckBox.Checked)
                {
                    stringBuilder.Append(explanationPair.Item2.AmountDeviation.ToString("G", culture));
                    stringBuilder.Append(this.separator);
                }
                if (this.percentCheckBox.Checked)
                {
                    double percent = 100.0 * (double)(explanationPair.Item2.Amount / explanation.Amount.Amount);
                    stringBuilder.Append(percent.ToString("G", culture));
                    stringBuilder.Append(this.separator);
                }
                stringBuilder.AppendLine();
            }
        }

        private string GetCSV()
        {
            StringBuilder stringBuilder = new StringBuilder();
            System.Globalization.CultureInfo culture;
            if (this.localRadioButton.Checked)
            {
                culture = System.Globalization.CultureInfo.CurrentCulture;
            }
            else
            {
                culture = System.Globalization.CultureInfo.InvariantCulture;
            }
            stringBuilder.Append("Name");
            stringBuilder.Append(this.separator);
            stringBuilder.Append("Amount");
            stringBuilder.Append(this.separator);
            if (this.deviationCheckBox.Checked)
            {
                stringBuilder.Append("Amount Deviation");
                stringBuilder.Append(this.separator);
            }
            if (this.percentCheckBox.Checked)
            {
                stringBuilder.Append("Percents");
                stringBuilder.Append(this.separator);
            }
            stringBuilder.AppendLine();
            if (this.IsInflowChecked)
            {
                this.BuildCSV(stringBuilder, this.periodInflowExplanation, "Total Inflow", culture);
            }
            if (this.IsInflowChecked && this.IsOutflowChecked)
            {
                stringBuilder.Append(this.separator);
                stringBuilder.Append(this.separator);
                if (this.deviationCheckBox.Checked)
                {
                    stringBuilder.Append(this.separator);
                }
                if (this.percentCheckBox.Checked)
                {
                    stringBuilder.Append(this.separator);
                }
                stringBuilder.AppendLine();
            }
            if (this.IsOutflowChecked)
            {
                this.BuildCSV(stringBuilder, this.periodOutflowExplanation, "Total Outflow", culture);
            }
            return stringBuilder.ToString();
        }

        private void exportButton_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "Csv files (*.csv)|*.csv|All files (*.*)|*.*";
            saveFileDialog.RestoreDirectory = true;
            System.Windows.Forms.DialogResult dialogResult = saveFileDialog.ShowDialog();
            if (dialogResult == System.Windows.Forms.DialogResult.OK)
            {
                File.WriteAllText(saveFileDialog.FileName, this.GetCSV());
            }
        }
    }
}
