﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

using CashCalc.Base;

namespace CashCalc.Dialogs
{
    /// <summary>
    /// Form for exporting the forecast time series diagram to the csv file.
    /// </summary>
    public partial class ExportSeriesToCsvForm : Form
    {
        /// <summary>
        /// Default constructor.
        /// </summary>
        public ExportSeriesToCsvForm()
        {
            InitializeComponent();
        }

        /// <summary>
        /// Constructor with specified forecast time series including 
        /// the corresponding inflow and outflow explanations.
        /// </summary>
        /// <param name="forecast">Forecast time series.</param>
        /// <param name="inflowForecastExplanation">Inflow explanation.</param>
        /// <param name="outflowForecastExplanation">Outflow explanation.</param>
        /// <param name="monthView"></param>
        public ExportSeriesToCsvForm(
            SortedList<DateTime, AmountWithDeviation> forecast,
            SortedList<DateTime, AmountExplanation> inflowForecastExplanation,
            SortedList<DateTime, AmountExplanation> outflowForecastExplanation,
            bool monthView)
            : this()
        {
            this.forecast = forecast;
            this.inflowForecastExplanation = inflowForecastExplanation;
            this.outflowForecastExplanation = outflowForecastExplanation;
            this.monthView = monthView;
        }

        private SortedList<DateTime, AmountWithDeviation> forecast;
        private SortedList<DateTime, AmountExplanation> inflowForecastExplanation;
        private SortedList<DateTime, AmountExplanation> outflowForecastExplanation;
        private bool monthView;

        private string separator = ";";

        private string GetCSV()
        {
            StringBuilder stringBuilder = new StringBuilder();
            System.Globalization.CultureInfo culture;
            if (this.localRadioButton.Checked)
            {
                culture = System.Globalization.CultureInfo.CurrentCulture;
            }
            else
            {
                culture = System.Globalization.CultureInfo.InvariantCulture;
            }
            if (this.dateCheckBox.Checked)
            {
                if (this.monthView)
                {
                    stringBuilder.Append("From Date");
                    stringBuilder.Append(this.separator);
                    stringBuilder.Append("To Date");
                    stringBuilder.Append(this.separator);
                }
                else
                {
                    stringBuilder.Append("Date");
                    stringBuilder.Append(this.separator);
                }
            }
            if (this.balanceCheckBox.Checked)
            {
                stringBuilder.Append("Balance");
                stringBuilder.Append(this.separator);
            }
            if (this.balanceDeviationCheckBox.Checked)
            {
                stringBuilder.Append("Balance Deviation");
                stringBuilder.Append(this.separator);
            }
            if (this.inflowVolumeCheckBox.Checked)
            {
                stringBuilder.Append("Inflow Volume");
                stringBuilder.Append(this.separator);
            }
            if (this.inflowVolumeDeviationCheckBox.Checked)
            {
                stringBuilder.Append("Inflow Volume Deviation");
                stringBuilder.Append(this.separator);
            }
            if (this.outflowVolumeCheckBox.Checked)
            {
                stringBuilder.Append("Outflow Volume");
                stringBuilder.Append(this.separator);
            }
            if (this.outflowVolumeDeviationCheckBox.Checked)
            {
                stringBuilder.Append("Outflow Volume Deviation");
                stringBuilder.Append(this.separator);
            }
            if (this.forecast != null && this.forecast.Count > 0)
            {
                foreach (KeyValuePair<DateTime, AmountWithDeviation> forecastPair in this.forecast)
                {
                    DateTime date = forecastPair.Key.Date;
                    AmountExplanation inflowExplanation = AmountExplanation.Zero;
                    AmountExplanation outflowExplanation = AmountExplanation.Zero;
                    if (this.inflowForecastExplanation != null && this.inflowForecastExplanation.ContainsKey(date))
                    {
                        inflowExplanation = this.inflowForecastExplanation[date];
                    }
                    if (this.outflowForecastExplanation != null && this.outflowForecastExplanation.ContainsKey(date))
                    {
                        outflowExplanation = this.outflowForecastExplanation[date];
                    }
                    stringBuilder.AppendLine();
                    if (this.dateCheckBox.Checked)
                    {
                        if (this.monthView)
                        {
                            stringBuilder.Append(date.ToString("d", culture));
                            stringBuilder.Append(this.separator);
                            stringBuilder.Append(date.AddMonths(1).AddTicks(-TimeSpan.TicksPerDay).ToString("d", culture));
                            stringBuilder.Append(this.separator);
                        }
                        else
                        {
                            stringBuilder.Append(date.ToString("d", culture));
                            stringBuilder.Append(this.separator);
                        }
                    }
                    if (this.balanceCheckBox.Checked)
                    {
                        stringBuilder.Append(forecastPair.Value.Amount.ToString("G", culture));
                        stringBuilder.Append(this.separator);
                    }
                    if (this.balanceDeviationCheckBox.Checked)
                    {
                        stringBuilder.Append(forecastPair.Value.AmountDeviation.ToString("G", culture));
                        stringBuilder.Append(this.separator);
                    }
                    if (this.inflowVolumeCheckBox.Checked)
                    {
                        stringBuilder.Append(inflowExplanation.Amount.Amount.ToString("G", culture));
                        stringBuilder.Append(this.separator);
                    }
                    if (this.inflowVolumeDeviationCheckBox.Checked)
                    {
                        stringBuilder.Append(inflowExplanation.Amount.AmountDeviation.ToString("G", culture));
                        stringBuilder.Append(this.separator);
                    }
                    if (this.outflowVolumeCheckBox.Checked)
                    {
                        stringBuilder.Append(outflowExplanation.Amount.Amount.ToString("G", culture));
                        stringBuilder.Append(this.separator);
                    }
                    if (this.outflowVolumeDeviationCheckBox.Checked)
                    {
                        stringBuilder.Append(outflowExplanation.Amount.AmountDeviation.ToString("G", culture));
                        stringBuilder.Append(this.separator);
                    }
                }
            }
            return stringBuilder.ToString();
        }

        private void exportButton_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.Filter = "Csv files (*.csv)|*.csv|All files (*.*)|*.*";
            saveFileDialog.RestoreDirectory = true;
            System.Windows.Forms.DialogResult dialogResult = saveFileDialog.ShowDialog();
            if (dialogResult == System.Windows.Forms.DialogResult.OK)
            {
                File.WriteAllText(saveFileDialog.FileName, this.GetCSV());
            }
        }
    }
}
