﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using CashCalc.Base;
using CashCalc.Common;
using CashCalc.Dialogs;

namespace CashCalc.UserControls
{
    /// <summary>
    /// User control for specifying any enumerable account.
    /// </summary>
    public partial class EnumerableAccountUserControl : UserControl
    {
        /// <summary>
        /// Default constructor.
        /// </summary>
        public EnumerableAccountUserControl()
        {
            InitializeComponent();
            this.Exceptions = new SerializableSortedList<DateTime, AmountWithDeviation>();
        }

        /// <summary>
        /// Delegate for obtaining the enumeration of valid dates.
        /// </summary>
        /// <returns>
        /// The enumeration of valid dates.
        /// </returns>
        public delegate IEnumerable<DateTime> ValidDatesHandler();

        /// <summary>
        /// Handler for obtaining the valid dates.
        /// </summary>
        public ValidDatesHandler ValidDates { get; set; }

        /// <summary>
        /// Enumerable account.
        /// </summary>
        public EnumerableAccount Account
        {
            set
            {
                this.Exceptions.Clear();
                this.StartDate = value.StartDate.Date;
                this.EndDate = value.EndDate.Date;
                this.Amount = value.Amount.Amount;
                this.AmountDeviation = value.Amount.AmountDeviation;
                if (value.Exceptions != null)
                {
                    this.Exceptions = value.Exceptions;
                    this.UpdateExceptionListView();
                }
            }
        }

        /// <summary>
        /// Starting date.
        /// </summary>
        public DateTime StartDate
        {
            get { return this.startDateTimePicker.Value.Date; }
            set { this.startDateTimePicker.Value = value.Date; }
        }

        /// <summary>
        /// Ending date.
        /// </summary>
        public DateTime EndDate
        {
            get
            {
                if (this.endDateCheckBox.Checked)
                {
                    return this.endDateTimePicker.Value.Date;
                }
                else
                {
                    return DateTime.MaxValue.Date;
                }
            }
            set
            {
                if (value.Date.Equals(DateTime.MaxValue.Date))
                {
                    this.endDateCheckBox.Checked = false;
                }
                else
                {
                    this.endDateCheckBox.Checked = true;
                    this.endDateTimePicker.Value = value.Date;
                }
            }
        }

        /// <summary>
        /// Amount of change of the account balance.
        /// </summary>
        public decimal Amount
        {
            get { return this.amountNumericUpDown.Value; }
            set { this.amountNumericUpDown.Value = value; }
        }

        /// <summary>
        /// Standard deviation of the amount.
        /// </summary>
        public double AmountDeviation
        {
            get { return (double)this.amountDeviationNumericUpDown.Value; }
            set { this.amountDeviationNumericUpDown.Value = (decimal)value; }
        }

        /// <summary>
        /// List of exceptions.
        /// </summary>
        public SerializableSortedList<DateTime, AmountWithDeviation> Exceptions { get; protected set; }

        /// <summary>
        /// Updates the list of exceptions.
        /// </summary>
        public void UpdateExceptionListView()
        {
            this.exceptionListView.BeginUpdate();
            this.exceptionListView.SelectedItems.Clear();
            this.exceptionListView.Items.Clear();

            if (this.Exceptions.Count > 0)
            {
                SortedList<DateTime, bool> validDatesMap = this.GetValidDatesMap(this.Exceptions.Keys);
                foreach (KeyValuePair<DateTime, AmountWithDeviation> exceptionPair in this.Exceptions)
                {
                    ListViewItem item = this.exceptionListView.Items.Add(
                        exceptionPair.Key.ToString("d", System.Globalization.CultureInfo.InvariantCulture));
                    item.Tag = exceptionPair.Key;
                    if (validDatesMap[exceptionPair.Key])
                    {
                        item.ForeColor = Color.Black;
                    }
                    else
                    {
                        item.ForeColor = Color.Gray;
                    }
                    item.SubItems.Add(exceptionPair.Value.Amount.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                    item.SubItems.Add(exceptionPair.Value.AmountDeviation.ToString("N", System.Globalization.CultureInfo.InvariantCulture));
                }
            }

            this.exceptionListView.EndUpdate();
        }

        private SortedList<DateTime, bool> GetValidDatesMap(IEnumerable<DateTime> dates)
        {
            SortedList<DateTime, bool> validDatesMap = new SortedList<DateTime, bool>();
            DateTime maxDate = DateTime.MinValue;
            bool isValidDatesNull = (this.ValidDates == null);
            foreach (DateTime date in dates)
            {
                validDatesMap[date] = isValidDatesNull;
                if (date > maxDate)
                {
                    maxDate = date;
                }
            }
            if (!isValidDatesNull)
            {
                IEnumerable<DateTime> validDates = this.ValidDates();
                foreach (DateTime validDate in validDates)
                {
                    if (validDate > maxDate)
                    {
                        break;
                    }
                    if (validDatesMap.ContainsKey(validDate))
                    {
                        validDatesMap[validDate] = true;
                    }
                }
            }
            return validDatesMap;
        }

        private void getMeanButton_Click(object sender, EventArgs e)
        {
            GetMeanForm getMeanForm = new GetMeanForm();
            if (getMeanForm.ShowDialog() == DialogResult.OK)
            {
                this.amountNumericUpDown.Value = (decimal)getMeanForm.Mean;
            }
        }

        private void getStdDevButton_Click(object sender, EventArgs e)
        {
            GetStdDevForm getStdDevForm = new GetStdDevForm();
            if (getStdDevForm.ShowDialog() == DialogResult.OK)
            {
                this.amountDeviationNumericUpDown.Value = (decimal)getStdDevForm.StdDev;
            }
        }

        private void startDateTimePicker_ValueChanged(object sender, EventArgs e)
        {
            this.endDateTimePicker.MinDate = this.startDateTimePicker.Value;
            this.UpdateExceptionListView();
        }

        private void endDateTimePicker_ValueChanged(object sender, EventArgs e)
        {
            this.UpdateExceptionListView();
        }

        private void endDateCheckBox_CheckedChanged(object sender, EventArgs e)
        {
            if (this.endDateCheckBox.Checked)
            {
                this.endDateTimePicker.Enabled = true;
                this.endDateTimePicker.MinDate = this.startDateTimePicker.Value;
            }
            else
            {
                this.endDateTimePicker.Enabled = false;
            }
            this.UpdateExceptionListView();
        }

        private void addExceptionButton_Click(object sender, EventArgs e)
        {
            DateAmountForm dateAmountForm = new DateAmountForm();
            dateAmountForm.MinDate = this.startDateTimePicker.Value;
            if (this.endDateCheckBox.Checked)
            {
                dateAmountForm.MaxDate = this.endDateTimePicker.Value;
            }
            if (dateAmountForm.ShowDialog() == DialogResult.OK)
            {
                if (this.Exceptions.ContainsKey(dateAmountForm.Date))
                {
                    if (MessageBox.Show("The exception list already contains item with date " +
                        dateAmountForm.Date.ToString("d", System.Globalization.CultureInfo.InvariantCulture) +
                        ". Do you want to replace it?", "Confirmation", MessageBoxButtons.YesNo) == DialogResult.Yes)
                    {
                        this.Exceptions[dateAmountForm.Date] = dateAmountForm.Amount;
                        this.UpdateExceptionListView();
                    }
                }
                else
                {
                    this.Exceptions[dateAmountForm.Date] = dateAmountForm.Amount;
                    this.UpdateExceptionListView();
                }
            }
        }

        private void removeExceptionButton_Click(object sender, EventArgs e)
        {
            if (this.exceptionListView.SelectedItems.Count > 0)
            {
                DateTime date = (DateTime)this.exceptionListView.SelectedItems[0].Tag;
                this.Exceptions.Remove(date);
                this.UpdateExceptionListView();
            }
        }

        private void editButton_Click(object sender, EventArgs e)
        {
            if (this.exceptionListView.SelectedItems.Count > 0)
            {
                DateTime date = (DateTime)this.exceptionListView.SelectedItems[0].Tag;
                AmountWithDeviation amount = this.Exceptions[date];
                DateAmountForm dateAmountForm = new DateAmountForm(date, amount.Amount, amount.AmountDeviation);
                if (dateAmountForm.ShowDialog() == DialogResult.OK)
                {
                    this.Exceptions[dateAmountForm.Date] = dateAmountForm.Amount;
                    this.UpdateExceptionListView();
                }
            }
        }

        private void clearButton_Click(object sender, EventArgs e)
        {
            if (this.Exceptions.Count > 0)
            {
                if (MessageBox.Show("Do you really want to clear the exception list?", 
                    "Confirmation", MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    this.Exceptions.Clear();
                    this.UpdateExceptionListView();
                }
            }
        }

        private void exceptionListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (this.exceptionListView.SelectedItems.Count > 0)
            {
                this.removeExceptionButton.Enabled = true;
                this.editButton.Enabled = true;
            }
            else
            {
                this.removeExceptionButton.Enabled = false;
                this.editButton.Enabled = false;
            }
        }
    }
}
